from synalinks.src.backend.common.json_schema_utils import standardize_schema
from synalinks.src.backend.common.name_scope import name_scope
from synalinks.src.backend.common.symbolic_data_model import SymbolicDataModel
from synalinks.src.backend.common.variables import Variable
from synalinks.src.backend.pydantic import core



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import numpy as np

from synalinks.src import testing
from synalinks.src.utils.plot_metrics import plot_metrics
from synalinks.src.utils.plot_metrics import plot_metrics_comparison
from synalinks.src.utils.plot_metrics import plot_metrics_comparison_with_mean_and_std
from synalinks.src.utils.plot_metrics import plot_metrics_with_mean_and_std


class PlotMetricsTest(testing.TestCase):
    def generate_random_metrics(self, metrics=None):
        """Generate random metrics values."""
        if metrics is None:
            metrics = ["reward", "reward_val"]
        return {metric: np.random.uniform(0.0, 1.0) for metric in metrics}

    def generate_random_metrics_list(self, metrics=None, num_runs=3):
        """Generate a list of random metrics for multiple runs."""
        return [self.generate_random_metrics(metrics) for _ in range(num_runs)]

    def generate_comparison_metrics(self, conditions=None, metrics=None):
        """Generate comparison metrics for multiple conditions."""
        if conditions is None:
            conditions = ["Program A", "Program B", "Program C"]
        if metrics is None:
            metrics = ["reward", "reward_val", "f1_score", "accuracy"]

        return {
            condition: self.generate_random_metrics(metrics) for condition in conditions
        }

    def generate_comparison_metrics_with_runs(
        self, conditions=None, metrics=None, num_runs=3
    ):
        """Generate comparison metrics with multiple runs for each condition."""
        if conditions is None:
            conditions = ["Program A", "Program B"]
        if metrics is None:
            metrics = ["reward", "reward_val", "f1_score", "accuracy"]

        return {
            condition: self.generate_random_metrics_list(metrics, num_runs)
            for condition in conditions
        }

    def test_plot_metrics(self):
        """Test basic plot_metrics function."""
        metrics = ["reward", "reward_val", "f1_score", "accuracy"]
        metrics = self.generate_random_metrics(metrics=metrics)
        plot_metrics(metrics, to_folder="/tmp/")

    def test_plot_metrics_with_mean_and_std(self):
        """Test plot_metrics_with_mean_and_std function."""
        metrics = ["reward", "reward_val", "f1_score", "accuracy"]
        metrics = self.generate_random_metrics(metrics=metrics)
        metrics1 = self.generate_random_metrics(metrics=metrics)
        plot_metrics_with_mean_and_std([metrics, metrics1], to_folder="/tmp/")

    def test_plot_metrics_comparison(self):
        """Test plot_metrics_comparison function."""
        metrics = ["reward", "reward_val", "f1_score", "accuracy"]
        conditions = ["Program A", "Program B"]
        comparison_metrics = self.generate_comparison_metrics(
            conditions=conditions, metrics=metrics
        )
        plot_metrics_comparison(comparison_metrics, to_folder="/tmp/")

    def test_plot_metrics_comparison_with_mean_and_std(self):
        """Test plot_metrics_comparison_with_mean_and_std function."""
        metrics = ["reward", "reward_val", "f1_score", "accuracy"]
        conditions = ["Program A", "Program B"]
        comparison_metrics = self.generate_comparison_metrics_with_runs(
            conditions=conditions, metrics=metrics, num_runs=5
        )
        plot_metrics_comparison_with_mean_and_std(comparison_metrics, to_folder="/tmp/")

    def test_plot_metrics_with_filter(self):
        """Test plot_metrics with metrics filter."""
        metrics = [
            "reward",
            "reward_val",
            "f1_score",
            "accuracy",
            "precision",
            "recall",
        ]
        metrics_dict = self.generate_random_metrics(metrics=metrics)
        metrics_filter = ["reward", "f1_score", "accuracy"]
        plot_metrics(metrics_dict, metrics_filter=metrics_filter, to_folder="/tmp/")

    def test_plot_metrics_comparison_with_filter(self):
        """Test plot_metrics_comparison with metrics filter."""
        metrics = [
            "reward",
            "reward_val",
            "f1_score",
            "accuracy",
            "precision",
            "recall",
        ]
        conditions = ["Program A", "Program B"]
        comparison_metrics = self.generate_comparison_metrics(
            conditions=conditions, metrics=metrics
        )
        metrics_filter = ["reward", "f1_score", "accuracy"]
        plot_metrics_comparison(
            comparison_metrics, metrics_filter=metrics_filter, to_folder="/tmp/"
        )

    def test_plot_metrics_with_high_values(self):
        """Test plot_metrics with values exceeding 1.0 to verify y-axis scaling."""
        metrics = {
            "reward": 0.85,
            "reward_val": 1.25,  # Value exceeding 1.0
            "f1_score": 0.92,
            "accuracy": 1.15,  # Value exceeding 1.0
        }
        plot_metrics(metrics, to_folder="/tmp/", to_file="test_high_values.png")

    def test_plot_metrics_comparison_with_high_values(self):
        """Test comparison plot with values exceeding 1.0."""
        comparison_metrics = {
            "Program A": {"reward": 0.85, "f1_score": 1.25, "accuracy": 0.92},
            "Program B": {"reward": 1.15, "f1_score": 0.88, "accuracy": 1.05},
        }
        plot_metrics_comparison(
            comparison_metrics,
            to_folder="/tmp/",
            to_file="test_comparison_high_values.png",
        )

    def test_plot_metrics_with_many_metrics(self):
        """Test plot_metrics with many metrics to verify x-axis rotation."""
        metrics = ["metric_" + str(i) for i in range(8)]  # 8 metrics to trigger rotation
        metrics_dict = self.generate_random_metrics(metrics=metrics)
        plot_metrics(metrics_dict, to_folder="/tmp/", to_file="test_many_metrics.png")

    def test_plot_metrics_comparison_with_show_values(self):
        """Test comparison plot with mean and std showing values on bars."""
        metrics = ["reward", "f1_score", "accuracy"]
        conditions = ["Program A", "Program B"]
        comparison_metrics = self.generate_comparison_metrics_with_runs(
            conditions=conditions, metrics=metrics, num_runs=4
        )
        plot_metrics_comparison_with_mean_and_std(
            comparison_metrics,
            show_values=True,
            to_folder="/tmp/",
            to_file="test_comparison_with_values.png",
        )



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.ops.json import concat as concat
from synalinks.src.ops.json import concat as concatenate
from synalinks.src.ops.json import factorize as factorize
from synalinks.src.ops.json import in_mask as in_mask
from synalinks.src.ops.json import logical_and as logical_and
from synalinks.src.ops.json import logical_or as logical_or
from synalinks.src.ops.json import logical_xor as logical_xor
from synalinks.src.ops.json import out_mask as out_mask
from synalinks.src.ops.json import prefix as prefix
from synalinks.src.ops.json import suffix as suffix
from synalinks.src.ops.language_models import predict as predict



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.modules import deserialize as deserialize
from synalinks.src.modules import serialize as serialize
from synalinks.src.modules.agents.function_calling_agent import (
    FunctionCallingAgent as FunctionCallingAgent,
)
from synalinks.src.modules.core.action import Action as Action
from synalinks.src.modules.core.branch import Branch as Branch
from synalinks.src.modules.core.decision import Decision as Decision
from synalinks.src.modules.core.generator import Generator as Generator
from synalinks.src.modules.core.identity import Identity as Identity
from synalinks.src.modules.core.input_module import Input as Input
from synalinks.src.modules.core.input_module import InputModule as InputModule
from synalinks.src.modules.core.not_module import Not as Not
from synalinks.src.modules.knowledge.embedding import Embedding as Embedding
from synalinks.src.modules.knowledge.update_knowledge import (
    UpdateKnowledge as UpdateKnowledge,
)
from synalinks.src.modules.merging.concat import Concat as Concat
from synalinks.src.modules.merging.concat import Concat as Concatenate
from synalinks.src.modules.merging.logical_and import And as And
from synalinks.src.modules.merging.logical_or import Or as Or
from synalinks.src.modules.merging.logical_xor import Xor as Xor
from synalinks.src.modules.module import Module as Module
from synalinks.src.modules.retrievers.entity_retriever import (
    EntityRetriever as EntityRetriever,
)
from synalinks.src.modules.retrievers.triplet_retriever import (
    TripletRetriever as TripletRetriever,
)
from synalinks.src.modules.synthesis.python_synthesis import (
    PythonSynthesis as PythonSynthesis,
)
from synalinks.src.modules.ttc.chain_of_thought import ChainOfThought as ChainOfThought
from synalinks.src.modules.ttc.self_critique import SelfCritique as SelfCritique



import inspect

from synalinks.src.api_export import synalinks_export
from synalinks.src.metrics.metric import Metric
from synalinks.src.metrics.reduction_metrics import Mean
from synalinks.src.metrics.reduction_metrics import MeanMetricWrapper
from synalinks.src.metrics.reduction_metrics import Sum
from synalinks.src.saving import serialization_lib
from synalinks.src.utils.naming import to_snake_case

ALL_OBJECTS = {
    # Base
    Metric,
    Mean,
    MeanMetricWrapper,
    Sum,
}

ALL_OBJECTS_DICT = {cls.__name__: cls for cls in ALL_OBJECTS}
ALL_OBJECTS_DICT.update({to_snake_case(cls.__name__): cls for cls in ALL_OBJECTS})


@synalinks_export("synalinks.metrics.serialize")
def serialize(metric):
    """Serializes metric function or `Metric` instance.

    Args:
        metric: A Synalinks `Metric` instance or a metric function.

    Returns:
        Metric configuration dictionary.
    """
    return serialization_lib.serialize_synalinks_object(metric)


@synalinks_export("synalinks.metrics.deserialize")
def deserialize(config, custom_objects=None):
    """Deserializes a serialized metric class/function instance.

    Args:
        config: Metric configuration.
        custom_objects: Optional dictionary mapping names (strings)
            to custom objects (classes and functions) to be
            considered during deserialization.

    Returns:
        A Synalinks `Metric` instance or a metric function.
    """
    return serialization_lib.deserialize_synalinks_object(
        config,
        module_objects=ALL_OBJECTS_DICT,
        custom_objects=custom_objects,
    )


@synalinks_export("synalinks.metrics.get")
def get(identifier):
    """Retrieves a Synalinks metric as a `function`/`Metric` class instance.

    The `identifier` may be the string name of a metric function or class.

    >>> metric = metrics.get("categorical_crossentropy")
    >>> type(metric)
    <class 'function'>
    >>> metric = metrics.get("CategoricalCrossentropy")
    >>> type(metric)
    <class '...metrics.CategoricalCrossentropy'>

    You can also specify `config` of the metric to this function by passing dict
    containing `class_name` and `config` as an identifier. Also note that the
    `class_name` must map to a `Metric` class

    >>> identifier = {"class_name": "CategoricalCrossentropy",
    ...               "config": {"from_logits": True}}
    >>> metric = metrics.get(identifier)
    >>> type(metric)
    <class '...metrics.CategoricalCrossentropy'>

    Args:
        identifier: A metric identifier. One of None or string name of a metric
            function/class or metric configuration dictionary or a metric
            function or a metric class instance

    Returns:
        A Synalinks metric as a `function`/ `Metric` class instance.
    """
    if identifier is None:
        return None
    if isinstance(identifier, dict):
        obj = deserialize(identifier)
    elif isinstance(identifier, str):
        obj = ALL_OBJECTS_DICT.get(identifier, None)
    else:
        obj = identifier
    if callable(obj):
        if inspect.isclass(obj):
            obj = obj()
        return obj
    else:
        raise ValueError(f"Could not interpret metric identifier: {identifier}")



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import tree
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import JsonDataModel
from synalinks.src.backend import SymbolicDataModel
from synalinks.src.modules.module import Module


@synalinks_export(["synalinks.modules.Not", "synalinks.Not"])
class Not(Module):
    """Not module.

    This module should be used as a placeholder when no operation is to be
    performed and the output should be None.

    This module is useful to implement stop conditions when combined with a conditional
    branch or as placeholder (like the Identity) before implementing guards that leverage
    the xor operation.

    Args:
        **kwargs (keyword arguments): The default module's arguments
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.built = True

    async def call(self, inputs):
        if isinstance(inputs, (JsonDataModel, SymbolicDataModel)):
            return None
        return tree.map_structure(
            lambda x: None,
            inputs,
        )

    async def compute_output_spec(self, inputs):
        if isinstance(inputs, (JsonDataModel, SymbolicDataModel)):
            return inputs.clone()
        return tree.map_structure(
            lambda x: x.clone(),
            inputs,
        )



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.modules.core.input_module import Input
from synalinks.src.modules.merging.logical_or import Or
from synalinks.src.programs.program import Program


class LogicalOrTest(testing.TestCase):
    async def test_logical_or_module_not_none(self):
        class Query(DataModel):
            query: str

        i0 = Input(data_model=Query)
        i1 = Input(data_model=Query)
        i2 = Input(data_model=Query)
        output = await Or()([i0, i1, i2])

        program = Program(
            inputs=[i0, i1, i2],
            outputs=output,
        )

        result = await program(
            [
                Query(query="a"),
                Query(query="b"),
                Query(query="c"),
            ]
        )

        expected_json = {
            "query": "a",
            "query_1": "b",
            "query_2": "c",
        }

        self.assertEqual(result.get_json(), expected_json)

    async def test_logical_or_module_none(self):
        class Query(DataModel):
            query: str

        i0 = Input(data_model=Query)
        i1 = Input(data_model=Query)
        i2 = Input(data_model=Query)
        output = await Or()([i0, i1, i2])

        program = Program(
            inputs=[i0, i1, i2],
            outputs=output,
        )

        result = await program(
            [
                None,
                Query(query="b"),
                None,
            ]
        )

        expected_json = {
            "query": "b",
        }

        self.assertEqual(result.get_json(), expected_json)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import re
import string

ARTICLE_REGEX = re.compile(r"\b(a|an|the|of|is)\b", re.UNICODE)
PUNCTUATION_TRANSLATOR = str.maketrans("", "", string.punctuation)

Y_ENDING = re.compile(r"[^aeiou]y$")
S_ENDING = re.compile(r"[sxz]$")
SH_CH_ENDING = re.compile(r"(sh|ch)$")
IES_ENDING = re.compile(r"ies$")
ES_ENDING = re.compile(r"es$")

SUFFIX_PATTERN = re.compile(r"_\d+$")

IRREGULAR_PLURALS = {
    "addendum": "addenda",
    "aircraft": "aircraft",
    "alga": "algae",
    "alumna": "alumnae",
    "alumnus": "alumni",
    "alveolus": "alveoli",
    "amoeba": "amoebae",
    "analysis": "analyses",
    "antenna": "antennae",
    "antithesis": "antitheses",
    "apex": "apices",
    "appendix": "appendices",
    "automaton": "automata",
    "axis": "axes",
    "bacillus": "bacilli",
    "bacterium": "bacteria",
    "baculum": "bacula",
    "barracks": "barracks",
    "basis": "bases",
    "beau": "beaux",
    "bison": "bison",
    "buffalo": "buffalo",
    "bureau": "bureaus",
    "cactus": "cacti",
    "calf": "calves",
    "carcinoma": "carcinomata",
    "carp": "carp",
    "census": "censuses",
    "chassis": "chassis",
    "cherub": "cherubim",
    "child": "children",
    "château": "châteaus",
    "cloaca": "cloacae",
    "cod": "cod",
    "codex": "codices",
    "concerto": "concerti",
    "consortium": "consortia",
    "corpus": "corpora",
    "crisis": "crises",
    "criterion": "criteria",
    "curriculum": "curricula",
    "cystoma": "cystomata",
    "datum": "data",
    "deer": "deer",
    "diagnosis": "diagnoses",
    "die": "dice",
    "dwarf": "dwarfs",
    "echo": "echoes",
    "elf": "elves",
    "elk": "elk",
    "ellipsis": "ellipses",
    "embargo": "embargoes",
    "emphasis": "emphases",
    "erratum": "errata",
    "faux pas": "faux pas",
    "fez": "fezes",
    "firmware": "firmware",
    "fish": "fish",
    "focus": "foci",
    "foot": "feet",
    "formula": "formulae",
    "fungus": "fungi",
    "gallows": "gallows",
    "genus": "genera",
    "glomerulus": "glomeruli",
    "goose": "geese",
    "graffito": "graffiti",
    "grouse": "grouse",
    "half": "halves",
    "hamulus": "hamuli",
    "hero": "heroes",
    "hippopotamus": "hippopotami",
    "hoof": "hooves",
    "hovercraft": "hovercraft",
    "hypothesis": "hypotheses",
    "iliac": "ilia",
    "incubus": "incubi",
    "index": "indices",
    "interstitium": "interstitia",
    "kakapo": "kakapo",
    "knife": "knives",
    "larva": "larvae",
    "leaf": "leaves",
    "libretto": "libretti",
    "life": "lives",
    "loaf": "loaves",
    "loculus": "loculi",
    "locus": "loci",
    "louse": "lice",
    "man": "men",
    "matrix": "matrices",
    "means": "means",
    "measles": "measles",
    "media": "media",
    "medium": "media",
    "memorandum": "memoranda",
    "millennium": "millennia",
    "minutia": "minutiae",
    "moose": "moose",
    "mouse": "mice",
    "nebula": "nebulae",
    "nemesis": "nemeses",
    "neurosis": "neuroses",
    "news": "news",
    "nucleolus": "nucleoli",
    "nucleus": "nuclei",
    "oasis": "oases",
    "occiput": "occipita",
    "offspring": "offspring",
    "omphalos": "omphaloi",
    "opus": "opera",
    "ovum": "ova",
    "ox": "oxen",
    "paralysis": "paralyses",
    "parenthesis": "parentheses",
    "person": "people",
    "phenomenon": "phenomena",
    "phylum": "phyla",
    "pike": "pike",
    "polyhedron": "polyhedra",
    "potato": "potatoes",
    "primus": "primi",
    "prognosis": "prognoses",
    "quiz": "quizzes",
    "radius": "radii",
    "referendum": "referenda",
    "salmon": "salmon",
    "scarf": "scarves",
    "scrotum": "scrota",
    "self": "selves",
    "seminoma": "seminomata",
    "series": "series",
    "sheep": "sheep",
    "shelf": "shelves",
    "shrimp": "shrimp",
    "simulacrum": "simulacra",
    "soliloquy": "soliloquies",
    "spacecraft": "spacecraft",
    "species": "species",
    "spectrum": "spectra",
    "squid": "squid",
    "stimulus": "stimuli",
    "stratum": "strata",
    "swine": "swine",
    "syconium": "syconia",
    "syllabus": "syllabi",
    "symposium": "symposia",
    "synopsis": "synopses",
    "synthesis": "syntheses",
    "tableau": "tableaus",
    "testis": "testes",
    "that": "those",
    "thesis": "theses",
    "thief": "thieves",
    "this": "these",
    "thrombus": "thrombi",
    "tomato": "tomatoes",
    "tooth": "teeth",
    "torus": "tori",
    "trout": "trout",
    "tuna": "tuna",
    "umbilicus": "umbilici",
    "uterus": "uteri",
    "vertebra": "vertebrae",
    "vertex": "vertices",
    "veto": "vetoes",
    "vita": "vitae",
    "vortex": "vortices",
    "watercraft": "watercraft",
    "wharf": "wharves",
    "wife": "wives",
    "wolf": "wolves",
    "woman": "women",
}

IRREGULAR_SINGULARS = {plural: singular for singular, plural in IRREGULAR_PLURALS.items()}


def to_plural(word):
    """
    Convert a singular word to its plural form.

    Args:
        word (str): The singular word to convert.

    Returns:
        (str): The plural form of the word.
    """
    if word in IRREGULAR_PLURALS:
        return IRREGULAR_PLURALS.get(word)
    else:
        # Use rules for regular plurals
        if Y_ENDING.search(word):
            return f"{word[:-1]}ies"
        elif S_ENDING.search(word) or SH_CH_ENDING.search(word):
            return f"{word}es"
        else:
            return f"{word}s"


def to_singular(word):
    """
    Convert a plural word to its singular form.

    Args:
        word (str): The plural word to convert.

    Returns:
        (str): The singular form of the word.
    """
    if word in IRREGULAR_SINGULARS:
        return IRREGULAR_SINGULARS.get(word)
    else:
        # Use rules for regular singulars
        if IES_ENDING.search(word):
            return f"{word[:-3]}y"
        elif ES_ENDING.search(word):
            if S_ENDING.search(word[:-2]) or SH_CH_ENDING.search(word[:-2]):
                return word[:-2]
            else:
                return word[:-1]
        elif word.endswith("s"):
            return word[:-1]
        else:
            return word


def to_plural_property(property_key):
    """
    Convert the last word of a property key to its plural form.

    Args:
        property_key (str): The property key to convert.

    Returns:
        (str): The property key with the last word in plural form.
    """
    words = property_key.split("_")
    if len(words) > 1:
        # Assume the last word is the noun
        words[-1] = to_plural(words[-1])
    else:
        words[0] = to_plural(words[0])
    return "_".join(words)


def to_singular_property(property_key):
    """
    Convert the last word of a property key to its singular form.

    Args:
        property_key (str): The property key to convert.

    Returns:
        (str): The property key with the last word in singular form.
    """
    words = property_key.split("_")
    if len(words) > 1:
        # Assume the last word is the noun
        words[-1] = to_singular(words[-1])
    else:
        words[0] = to_singular(words[0])
    return "_".join(words)


def remove_numerical_suffix(property_key):
    """
    Remove the numerical suffix from a property key.

    Args:
        property_key (str): The property key to process.

    Returns:
        (str): The property key with the suffix removed.
    """
    return re.sub(SUFFIX_PATTERN, "", property_key)


def add_suffix(property_key, suffix):
    """
    Add a suffix to a property key.

    Args:
        property_key (str): The property key to process.
        suffix (int): The suffix to add.

    Returns:
        (str): The property key with the suffix added.
    """
    return f"{property_key}_{suffix}"


def to_singular_without_numerical_suffix(property_key):
    """
    Convert a property key to its base (singular) form by removing
        the numerical suffix and converting to singular.

    Args:
        property_key (str): The property key to convert.

    Returns:
        (str): The base (singular) form of the property key.
    """
    property_key = remove_numerical_suffix(property_key)
    return to_singular_property(property_key)


def to_plural_without_numerical_suffix(property_key):
    """
    Convert a property key to its list (plural) form by removing
        the numerical suffix and converting to plural.

    Args:
        property_key (str): The property key to convert.

    Returns:
        (str): The list (plural) form of the property key.
    """
    property_key = remove_numerical_suffix(property_key)
    return to_plural_property(property_key)


def is_plural(property_key):
    """
    Check if the last word of a property key is in plural form.

    Args:
        property_key (str): The property key to check.

    Returns:
        (bool): True if the last word is plural, False otherwise.
    """
    words = property_key.split("_")
    if len(words) > 1:
        noun = words[-1]
    else:
        noun = words[0]

    singular_form = to_singular(noun)
    return singular_form != noun


def remove_articles(text):
    """
    Remove common English articles from the text.

    Args:
        text (str): The text to process.

    Returns:
        (str): The text with articles removed.
    """
    return " ".join(re.sub(ARTICLE_REGEX, "", text).split())


def remove_punctuation(text):
    """
    Remove punctuation from the text.

    Args:
        text (str): The text to process.

    Returns:
        (str): The text with punctuation removed.
    """
    return text.translate(PUNCTUATION_TRANSLATOR)


def normalize_text(text):
    """
    Normalize the text by converting to lowercase, removing articles,
        and removing punctuation.

    Args:
        text (str): The text to normalize.

    Returns:
        (str): The normalized text.
    """
    return remove_articles(remove_punctuation(text.strip().lower()))


def normalize_and_tokenize(text):
    """
    Normalize the text and tokenize it into words.

    Args:
        text (str): The text to process.

    Returns:
        (list): A list of normalized words.
    """
    text = text.lower()
    text = remove_articles(text)
    text = remove_punctuation(text)
    return text.split()


def shorten_text(text, nb_words_offset=10):
    """
    Shorten a text.

    Args:
        text (str): The text to shorten.
        nb_words_offset (int): The number of words to keep
            from the beginning and end of the text
            (Default is 20).

    Returns:
        (str): The shortened text. If the original text has more than nb_words words,
            returns the first nb_words and last nb_words separated by " (...) ".
            Otherwise, returns the original text unchanged.
    """
    if not isinstance(text, str):
        text = str(text)

    words = text.split(" ")
    if len(words) <= nb_words_offset * 2:
        return text
    nb_words_removed = len(words) - 2 * nb_words_offset
    short_text = (
        " ".join(words[:nb_words_offset])
        + f" (... {nb_words_removed} words removed for clarity) "
        + " ".join(words[-nb_words_offset:])
    )
    return short_text



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import json
from unittest.mock import patch

from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.backend import Field
from synalinks.src.language_models import LanguageModel
from synalinks.src.modules.core.input_module import Input
from synalinks.src.modules.ttc.chain_of_thought import ChainOfThought
from synalinks.src.programs.program import Program


class ChainOfThoughtModuleTest(testing.TestCase):
    @patch("litellm.acompletion")
    async def test_chain_of_thought_with_k_1(self, mock_completion):
        class Query(DataModel):
            query: str = Field(
                description="The user query",
            )

        class Answer(DataModel):
            answer: str = Field(
                description="The correct answer",
            )

        language_model = LanguageModel(
            model="ollama/mistral",
        )

        x0 = Input(data_model=Query)
        x1 = await ChainOfThought(
            data_model=Answer,
            language_model=language_model,
        )(x0)

        program = Program(
            inputs=x0,
            outputs=x1,
            name="chain_of_tought_qa",
            description="Answer the user query step by step",
        )

        expected_string = (
            """{"thinking": "Toulouse hosts numerous research institutions """
            """and universities that specialize in aerospace engineering and """
            """robotics, such as the Institut Supérieur de l'Aéronautique et """
            """de l'Espace (ISAE-SUPAERO) and the French National Centre for """
            """Scientific Research (CNRS)","""
            """ "answer": "Toulouse"}"""
        )

        mock_completion.return_value = {
            "choices": [{"message": {"content": expected_string}}]
        }

        result = await program(
            Query(
                query="What is the French city of aerospace and robotics?",
            )
        )

        self.assertEqual(result.get_json(), json.loads(expected_string))

    @patch("litellm.acompletion")
    async def test_chain_of_thought_with_k_2(self, mock_completion):
        class Query(DataModel):
            query: str = Field(
                description="The user query",
            )

        class Answer(DataModel):
            answer: str = Field(
                description="The correct answer",
            )

        language_model = LanguageModel(
            model="ollama_chat/deepseek-r1",
        )

        x0 = Input(data_model=Query)
        x1 = await ChainOfThought(
            data_model=Answer,
            language_model=language_model,
            k=2,
        )(x0)

        program = Program(
            inputs=x0,
            outputs=x1,
            name="chain_of_tought_qa",
            description="Answer the user query step by step",
        )

        expected_string = (
            """{"thinking": "Toulouse hosts numerous research institutions """
            """and universities that specialize in aerospace engineering and """
            """robotics, such as the Institut Supérieur de l'Aéronautique et """
            """de l'Espace (ISAE-SUPAERO) and the French National Centre for """
            """Scientific Research (CNRS)","""
            """ "thinking_1": "Also Toulouse is the city where the Airbus """
            """planes are being assembled","""
            """ "answer": "Toulouse"}"""
        )

        mock_completion.return_value = {
            "choices": [{"message": {"content": expected_string}}]
        }

        result = await program(
            Query(
                query="What is the French city of aerospace and robotics?",
            )
        )

        self.assertEqual(result.get_json(), json.loads(expected_string))



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from unittest.mock import patch

from synalinks.src import testing
from synalinks.src.language_models import LanguageModel
from synalinks.src.modules import Generator
from synalinks.src.modules import Input
from synalinks.src.optimizers.random_few_shot import RandomFewShot
from synalinks.src.programs import Program
from synalinks.src.rewards.exact_match import ExactMatch
from synalinks.src.testing.test_utils import AnswerWithRationale
from synalinks.src.testing.test_utils import Query
from synalinks.src.testing.test_utils import load_test_data
from synalinks.src.testing.test_utils import mock_completion_data
from synalinks.src.testing.test_utils import mock_incorrect_completion_data


class RandomFewShotTest(testing.TestCase):
    @patch("litellm.acompletion")
    async def test_random_few_shot_training(self, mock_completion):
        language_model = LanguageModel(
            model="ollama/mistral",
        )

        inputs = Input(data_model=Query)
        outputs = await Generator(
            language_model=language_model,
            data_model=AnswerWithRationale,
        )(inputs)

        program = Program(
            inputs=inputs,
            outputs=outputs,
            name="test_program",
            description="A test program",
        )

        program.compile(
            optimizer=RandomFewShot(
                nb_min_examples=1,
            ),
            reward=ExactMatch(in_mask=["answer"]),
        )

        (x_train, y_train), (x_test, y_test) = load_test_data()

        mock_responses = []
        mock_responses.extend(mock_incorrect_completion_data())
        mock_responses.extend(mock_completion_data())

        mock_completion.side_effect = mock_responses

        _ = await program.fit(
            x=x_train,
            y=y_train,
            epochs=2,
            batch_size=32,
        )

        program_vars = program.get_variable(index=0).get_json()
        self.assertTrue(len(program_vars["examples"]) > 0)



# Modified from: keras/src/utils/summary_utils.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import copy
import json
import re
import shutil

import rich
import rich.table

from synalinks.src import tree
from synalinks.src.utils import io_utils


def count_params(variables):
    return int(sum(len(v.get_json()) for v in variables))


def highlight_number(x):
    """Themes numbers in a summary using rich markup.

    We use a separate color for `None`s, e.g. in a module schema.
    """
    if x is None:
        return f"[color(45)]{x}[/]"
    else:
        return f"[color(34)]{x}[/]"


def highlight_symbol(x):
    """Themes keras symbols in a summary using rich markup."""
    return f"[color(33)]{x}[/]"


def bold_text(x, color=None):
    """Bolds text using rich markup."""
    if color:
        return f"[bold][color({color})]{x}[/][/]"
    return f"[bold]{x}[/]"


def format_module_schema(module):
    if not module._inbound_nodes and not module._build_schemas_dict:
        return "?"

    def format_schema(schema):
        schema = copy.deepcopy(schema)
        schema.pop("$defs")
        return json.dumps(schema, indent=2)

    # There are 2 approaches to get output schemas:
    # 1. Using `module._inbound_nodes`, which is possible if the program is a
    # Sequential or Functional.
    # 2. Using `module._build_schemas_dict`, which is possible if users manually
    # build the module.
    if len(module._inbound_nodes) > 0:
        for i in range(len(module._inbound_nodes)):
            outputs = module._inbound_nodes[i].output_data_models
            output_schemas = tree.map_structure(
                lambda x: format_schema(x.get_schema()), outputs
            )
    else:
        try:
            if hasattr(module, "output_schema"):
                output_schemas = format_schema(module.output_schema)
            else:
                outputs = module.compute_output_schema(**module._build_schemas_dict)
                output_schemas = tree.map_schema_structure(
                    lambda x: format_schema(x), outputs
                )
        except NotImplementedError:
            return "?"
    if len(output_schemas) == 1:
        return output_schemas[0]
    out = "\n---\n".join(output_schemas)
    return out


def print_summary(
    program,
    line_length=None,
    positions=None,
    print_fn=None,
    expand_nested=False,
    show_trainable=False,
    module_range=None,
):
    """Prints a summary of a program.

    Args:
        program: Synalinks program instance.
        line_length: Total length of printed lines
            (e.g. set this to adapt the display to different
            terminal window sizes).
        positions: Relative or absolute positions of log elements in each line.
            If not provided, defaults to `[0.3, 0.6, 0.70, 1.]`.
        print_fn: Print function to use.
            It will be called on each line of the summary.
            You can set it to a custom function
            in order to capture the string summary.
            It defaults to `print` (prints to stdout).
        return_string: If True, return the summary string instead of printing.
        expand_nested: Whether to expand the nested programs.
            If not provided, defaults to `False`.
        show_trainable: Whether to show if a module is trainable.
            If not provided, defaults to `False`.
        module_range: List or tuple containing two strings,
            the starting module name and ending module name (both inclusive),
            indicating the range of modules to be printed in the summary. The
            strings could also be regexes instead of an exact name. In this
             case, the starting module will be the first module that matches
            `module_range[0]` and the ending module will be the last element that
            matches `module_range[1]`. By default (`None`) all
            modules in the program are included in the summary.
    """
    from synalinks.src.programs import Functional
    from synalinks.src.programs import Sequential

    if not print_fn and not io_utils.is_interactive_logging_enabled():
        print_fn = io_utils.print_msg

    if isinstance(program, Sequential):
        sequential_like = True
        modules = program.modules
    elif not isinstance(program, Functional):
        # We treat subclassed programs as a simple sequence of modules, for logging
        # purposes.
        sequential_like = True
        modules = program.modules
    else:
        modules = program._operations
        sequential_like = True
        nodes_by_depth = program._nodes_by_depth.values()
        nodes = []
        for v in nodes_by_depth:
            if (len(v) > 1) or (
                len(v) == 1 and len(tree.flatten(v[0].input_data_models)) > 1
            ):
                # if the program has multiple nodes
                # or if the nodes have multiple inbound_modules
                # the program is no longer sequential
                sequential_like = False
                break
            nodes += v
        if sequential_like:
            # search for shared modules
            for module in program.modules:
                flag = False
                for node in module._inbound_nodes:
                    if node in nodes:
                        if flag:
                            sequential_like = False
                            break
                        else:
                            flag = True
                if not sequential_like:
                    break

    if sequential_like:
        default_line_length = 88
        positions = positions or [0.45, 0.80, 1.0]
        # header names for the different log elements
        header = ["Module (type)", "Output Schema", "Variable #"]
        alignment = ["left", "left", "right"]
    else:
        default_line_length = 108
        positions = positions or [0.3, 0.56, 0.74, 1.0]
        # header names for the different log elements
        header = [
            "Module (type)",
            "Output Schema",
            "Variable #",
            "Connected to",
        ]
        alignment = ["left", "left", "right", "left"]
        relevant_nodes = []
        for v in program._nodes_by_depth.values():
            relevant_nodes += v

    if show_trainable:
        default_line_length += 12
        positions = [p * 0.90 for p in positions] + [1.0]
        header.append("Trainable")
        alignment.append("center")

    # Compute columns widths
    default_line_length = min(default_line_length, shutil.get_terminal_size().columns - 4)
    line_length = line_length or default_line_length
    column_widths = []
    current = 0
    for pos in positions:
        width = int(pos * line_length) - current
        if width < 4:
            raise ValueError("Insufficient console width to print summary.")
        column_widths.append(width)
        current += width

    # Render summary as a rich table.
    columns = []
    # Right align parameter counts.
    for i, name in enumerate(header):
        column = rich.table.Column(
            name,
            justify=alignment[i],
            width=column_widths[i],
        )
        columns.append(column)

    table = rich.table.Table(*columns, width=line_length, show_lines=True)

    def get_connections(module):
        connections = ""
        for node in module._inbound_nodes:
            if relevant_nodes and node not in relevant_nodes:
                # node is not part of the current network
                continue
            for dm in node.input_data_models:
                synalinks_history = dm._synalinks_history
                inbound_module = synalinks_history.operation
                node_index = highlight_number(synalinks_history.node_index)
                data_model_index = highlight_number(synalinks_history.data_model_index)
                if connections:
                    connections += ", "
                connections += f"{inbound_module.name}[{node_index}][{data_model_index}]"
        if not connections:
            connections = "-"
        return connections

    def get_module_fields(module, prefix=""):
        output_schema = format_module_schema(module)
        name = prefix + module.name
        cls_name = module.__class__.__name__
        name = rich.markup.escape(name)
        name += f" ({highlight_symbol(rich.markup.escape(cls_name))})"

        if not hasattr(module, "built"):
            params = highlight_number(0)
        elif not module.built:
            params = highlight_number(0) + " (unbuilt)"
        else:
            params = highlight_number(f"{count_params(module.variables):,}")

        fields = [name, output_schema, params]
        if not sequential_like:
            fields.append(get_connections(module))
        if show_trainable:
            if hasattr(module, "variables") and len(module.variables) > 0:
                fields.append(
                    bold_text("Y", color=34)
                    if module.trainable
                    else bold_text("N", color=9)
                )
            else:
                fields.append(bold_text("-"))
        return fields

    def print_module(module, nested_level=0):
        if nested_level:
            prefix = "   " * nested_level + "└" + " "
        else:
            prefix = ""

        fields = get_module_fields(module, prefix=prefix)

        rows = [fields]
        if expand_nested and hasattr(module, "modules") and module.modules:
            nested_modules = module.modules
            nested_level += 1
            for i in range(len(nested_modules)):
                rows.extend(print_module(nested_modules[i], nested_level=nested_level))
        return rows

    # Render all modules to the rich table.
    module_range = get_module_index_bound_by_module_name(modules, module_range)
    for module in modules[module_range[0] : module_range[1]]:
        for row in print_module(module):
            table.add_row(*row)

    # Create a rich console for printing. Capture for non-interactive logging.
    if print_fn:
        console = rich.console.Console(
            highlight=False, force_terminal=False, color_system=None
        )
        console.begin_capture()
    else:
        console = rich.console.Console(highlight=False)

    # Print the to the console.
    console.print(f"Program: {rich.markup.escape(program.name)}")
    console.print(f"description: '{rich.markup.escape(program.description)}'")
    console.print(table)


def get_module_index_bound_by_module_name(modules, module_range=None):
    """Get the module indexes from the model based on module names.

    The module indexes can be used to slice the model into sub models for
    display.

    Args:
        model: `Model` instance.
        module_names: a list or tuple of 2 strings, the starting module name and
            ending module name (both inclusive) for the result. All modules will
            be included when `None` is provided.

    Returns:
        The index value of module based on its unique name (module_names).
        Output will be [first_module_index, last_module_index + 1].
    """
    if module_range is not None:
        if len(module_range) != 2:
            raise ValueError(
                "module_range must be a list or tuple of length 2. Received: "
                f"module_range = {module_range} of length {len(module_range)}"
            )
        if not isinstance(module_range[0], str) or not isinstance(module_range[1], str):
            raise ValueError(
                f"module_range should contain string type only. Received: {module_range}"
            )
    else:
        return [0, len(modules)]

    lower_index = [
        idx
        for idx, module in enumerate(modules)
        if re.match(module_range[0], module.name)
    ]
    upper_index = [
        idx
        for idx, module in enumerate(modules)
        if re.match(module_range[1], module.name)
    ]

    if not lower_index or not upper_index:
        raise ValueError(
            "Passed module_names do not match the module names in the model. "
            f"Received: {module_range}"
        )

    if min(lower_index) > max(upper_index):
        return [min(upper_index), max(lower_index) + 1]
    return [min(lower_index), max(upper_index) + 1]



import types

from synalinks.src.trainers.data_adapters import array_data_adapter
from synalinks.src.trainers.data_adapters import data_adapter
from synalinks.src.trainers.data_adapters.array_data_adapter import ArrayDataAdapter
from synalinks.src.trainers.data_adapters.generator_data_adapter import (
    GeneratorDataAdapter,
)


def get_data_adapter(
    x,
    y=None,
    batch_size=None,
    steps_per_epoch=None,
    shuffle=False,
):
    # Allow passing a custom data adapter.
    if isinstance(x, data_adapter.DataAdapter):
        return x

    if array_data_adapter.can_convert_arrays((x, y)):
        return ArrayDataAdapter(
            x,
            y,
            shuffle=shuffle,
            batch_size=batch_size,
            steps=steps_per_epoch,
        )
    elif isinstance(x, types.GeneratorType):
        if y is not None:
            raise_unsupported_arg("y", "the targets", "PyDataset")
        return GeneratorDataAdapter(x)

    else:
        raise ValueError(f"Unrecognized data type: x={x} (of type {type(x)})")


def raise_unsupported_arg(arg_name, arg_description, input_type):
    raise ValueError(
        f"When providing `x` as a {input_type}, `{arg_name}` "
        f"should not be passed. Instead, {arg_description} should "
        f"be included as part of the {input_type}."
    )



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import json
from unittest.mock import patch

from synalinks.src import modules
from synalinks.src import testing
from synalinks.src.backend import DataModel
from synalinks.src.backend import SymbolicDataModel
from synalinks.src.backend import standardize_schema
from synalinks.src.language_models import LanguageModel
from synalinks.src.modules import Input
from synalinks.src.programs import Sequential


class SequentialTest(testing.TestCase):
    @patch("litellm.acompletion")
    async def test_basic_flow_with_input(self, mock_completion):
        class Query(DataModel):
            query: str

        class AnswerWithRationale(DataModel):
            rationale: str
            answer: str

        language_model = LanguageModel(
            model="ollama_chat/deepseek-r1",
        )

        expected_string = (
            """{"rationale": "Toulouse hosts numerous research institutions """
            """and universities that specialize in aerospace engineering and """
            """robotics, such as the Institut Supérieur de l'Aéronautique et """
            """de l'Espace (ISAE-SUPAERO) and the French National Centre for """
            """Scientific Research (CNRS)","""
            """ "answer": "Toulouse"}"""
        )

        mock_completion.return_value = {
            "choices": [{"message": {"content": expected_string}}]
        }

        program = Sequential(
            name="chain_of_thought",
            description="Useful to answer in a step by step manner",
        )
        program.add(Input(data_model=Query))
        program.add(
            modules.Generator(
                data_model=AnswerWithRationale,
                language_model=language_model,
            )
        )

        self.assertEqual(len(program.modules), 2)
        self.assertTrue(program.built)
        self.assertEqual(len(program.variables), 1)

        # Test eager call
        result = await program(
            Query(query="What is the french city of aerospace and robotics?")
        )
        self.assertEqual(result.get_json(), json.loads(expected_string))

        # Test symbolic call
        x = SymbolicDataModel(data_model=Query)
        y = await program(x)
        self.assertIsInstance(y, SymbolicDataModel)
        self.assertEqual(
            y.get_schema(), standardize_schema(AnswerWithRationale.get_schema())
        )

        # Test `modules` constructor arg
        program = Sequential(
            modules=[
                Input(data_model=Query),
                modules.Generator(
                    data_model=AnswerWithRationale,
                    language_model=language_model,
                ),
            ],
            name="chain_of_thought",
            description="Useful to answer in a step by step manner",
        )
        self.assertEqual(len(program.modules), 2)
        self.assertTrue(program.built)
        self.assertEqual(len(program.variables), 1)

        result = await program(
            Query(query="What is the french city of aerospace and robotics?")
        )
        self.assertEqual(result.get_json(), json.loads(expected_string))

        # Test pop
        program.pop()
        self.assertEqual(len(program.modules), 1)
        self.assertFalse(program.built)
        self.assertEqual(len(program.variables), 0)

    def test_representation(self):
        program = Sequential(
            name="chain_of_thought",
            description="Useful to answer in a step by step manner",
        )
        self.assertEqual(
            str(program),
            "<Sequential name=chain_of_thought, "
            "description='Useful to answer in a step by step manner', "
            "built=False>",
        )



# MIT License
# Copyright (c) 2023 Lucas Lofaro

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Modified from: llmfuncs
# Original authors: Lucas Lofaro
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import inspect
import logging
import typing

import docstring_parser
from tenacity import retry
from tenacity import retry_if_exception_type
from tenacity import stop_after_attempt
from tenacity import wait_exponential

from synalinks.src.api_export import synalinks_export
from synalinks.src.saving import serialization_lib
from synalinks.src.saving.synalinks_saveable import SynalinksSaveable

JsonSchema = typing.Union[
    typing.Dict[str, typing.Any],
    typing.List[typing.Any],
    str,
    int,
    float,
    bool,
    None,
]


def json_schema_type(py_type: typing.Any) -> JsonSchema:
    mapping = {
        int: "integer",
        float: "number",
        bool: "boolean",
        str: "string",
        type(None): "null",
    }

    # Check if type is a basic type
    if py_type in mapping:
        return mapping[py_type]

    # For unparameterized list and dict types
    if py_type is list or py_type is typing.List:
        return {"type": "array", "items": {}}
    if py_type is dict or py_type is typing.Dict:
        return {"type": "object", "additionalProperties": {}}

    origin = typing.get_origin(py_type)
    args = typing.get_args(py_type)

    if origin is typing.Union:
        # this is a special case to handle Optional[type] which is just syntactic sugar
        # around Union[type, None]
        if len(args) == 2 and type(None) in args:
            # Assuming the None is always last
            return json_schema_type(args[0])
        else:
            return [json_schema_type(arg) for arg in args]

    if origin is list or origin is typing.List:
        # For simplicity, we're assuming all elements in the list are of the same type
        schema_type = json_schema_type(args[0])
        if isinstance(schema_type, dict):
            return {"type": "array", "items": schema_type}
        return {
            "type": "array",
            "items": {"type": json_schema_type(args[0])},
        }

    if origin is dict or origin is typing.Dict:
        # For simplicity, we're assuming all keys are strings
        # and all values are of the same type
        schema_type = json_schema_type(args[1])
        if isinstance(schema_type, dict):
            return {"type": "object", "additionalProperties": schema_type}
        return {"type": "object", "additionalProperties": {"type": schema_type}}

    # The type is not supported
    raise ValueError(f"Cannot convert {py_type} to a JSON schema type")


def get_param_schema(
    param_name: str,
    param: inspect.Parameter,
    type_hints: typing.Dict[str, typing.Any],
    doc_parsed: docstring_parser.Docstring,
) -> JsonSchema:
    """Create a schema for a single parameter."""
    if param_name not in type_hints:
        raise ValueError(f"Missing type hint for parameter '{param_name}'")
    param_type = type_hints[param_name]
    param_type_str = json_schema_type(param_type)
    descriptions = (p.description for p in doc_parsed.params if p.arg_name == param_name)
    param_doc = next(descriptions, None)
    if param_doc is None:
        raise ValueError(f"Missing description for parameter '{param_name}' in docstring")

    param_schema = {}
    param_schema["description"] = param_doc.replace("\n", " ")
    param_schema["title"] = param_name.title().replace("_", " ")
    # Check if the param_type_str is already a dictionary.
    if isinstance(param_type_str, dict):
        param_schema.update(**param_type_str)
    else:
        param_schema["type"] = param_type_str

    if param.default is not param.empty:
        param_schema["default"] = param.default

    return param_schema


@synalinks_export(
    [
        "synalinks.utils.Tool",
        "synalinks.Tool",
    ]
)
class Tool(SynalinksSaveable):
    def __init__(self, func: typing.Callable):
        self._func = func
        if not inspect.iscoroutinefunction(self._func):
            raise TypeError(f"{self.name} is not an asynchronous function")

        doc = inspect.getdoc(func)
        if not doc:
            raise ValueError(f"The tool ({self.name}) must have a docstring")

        self._docstring = docstring_parser.parse(doc)
        self._signature = inspect.signature(func)
        self._type_hints = typing.get_type_hints(func)
        self._params_schema = {}
        self._required_params = []

        self._parse_arguments()

        if not self.description:
            logging.warning(
                "The tool (%s) has no description. " % self.name
                + "This is unsafe behavior and may lead to issues."
            )

    @retry(
        stop=stop_after_attempt(3),
        wait=wait_exponential(multiplier=1, min=1, max=10),
        retry=retry_if_exception_type((Exception,)),
        reraise=True,
    )
    async def __call__(self, *args, **kwargs):
        return await self._func(*args, **kwargs)

    def _parse_arguments(self):
        for param_name, param in self._signature.parameters.items():
            param_schema = get_param_schema(
                param_name,
                param,
                self._type_hints,
                self._docstring,
            )
            self._params_schema[param_name] = param_schema
            if param.default is param.empty:
                self._required_params.append(param_name)

    @property
    def description(self) -> str:
        return self._docstring.short_description or ""

    @property
    def name(self) -> str:
        return self._func.__name__

    def get_tool_schema(self):
        schema = {
            "additionalProperties": False,
            "description": self._docstring.short_description,
            "properties": self._params_schema,
            "required": self._required_params,
            "title": self.name.title().replace("_", " "),
            "type": "object",
        }
        return schema

    def get_config(self):
        func_config = {"func": serialization_lib.serialize_synalinks_object(self._func)}
        return {**func_config}

    @classmethod
    def from_config(cls, config):
        func = serialization_lib.deserialize_synalinks_object(config.pop("func"))
        return cls(func)



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.optimizers.omega import OMEGA as OMEGA
from synalinks.src.optimizers.random_few_shot import RandomFewShot as RandomFewShot



from synalinks.src.utils.python_utils import default
from synalinks.src.utils.python_utils import is_default
from synalinks.src.utils.python_utils import removeprefix
from synalinks.src.utils.python_utils import removesuffix



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import copy
import re
from typing import Any
from typing import Dict

from synalinks.src.embedding_models import EmbeddingModel
from synalinks.src.utils.async_utils import run_maybe_nested
from synalinks.src.utils.naming import to_snake_case


class DatabaseAdapter:
    def __init__(
        self,
        uri=None,
        embedding_model=None,
        entity_models=None,
        relation_models=None,
        metric="cosine",
        wipe_on_start=False,
        **kwargs,
    ):
        self.uri = uri
        if not embedding_model:
            raise ValueError("KnowledgeBase requires `embedding_model` argument")
        if not isinstance(embedding_model, EmbeddingModel):
            raise ValueError(
                "KnowledgeBase `embedding_model` argument to be an `EmbeddingModel`"
            )

        self.embedding_model = embedding_model
        self.embedding_dim = len(
            run_maybe_nested(embedding_model(texts=["test"]))["embeddings"][0]
        )

        if metric not in ("cosine", "euclidean"):
            raise ValueError(
                "KnowledgeBase `metric` argument should between `cosine` or `euclidean`"
            )
        self.metric = metric

        if not entity_models and not relation_models:
            raise ValueError(
                "KnowledgeBase requires `entity_models` and `relation_models` argument"
            )
        if not entity_models:
            raise ValueError("KnowledgeBase requires `entity_models` argument")

        if not relation_models:
            relation_models = []

        self.entity_models = entity_models
        self.relation_models = relation_models

        if wipe_on_start:
            self.wipe_database()

        self.create_vector_index()

    def sanitize(self, string):
        """Prevent Cypher injections.

        Args:
            string (str): The string to sanitize

        Returns:
            (str): The sanitized string
        """
        string = re.sub(r"//.*?(\n|$)", "", string)
        string = re.sub(r"/\*.*?\*/", "", string, flags=re.DOTALL)
        return string

    def sanitize_label(self, label):
        """Prevent Cypher injections.

        Args:
            label (str): The label to sanitize

        Returns:
            (str): The sanitized string
        """
        label = self.sanitize(label)
        label = re.sub(r"[^a-zA-Z0-9]", "", label)
        return label

    def sanitize_property_name(self, property_name):
        """Prevent Cypher injections.

        Args:
            property_name (str): The property name to sanitize

        Returns:
            (str): The sanitized property name
        """
        property_name = self.sanitize(property_name)
        property_name = re.sub(r"[^a-zA-Z0-9_]", "", property_name)
        return to_snake_case(property_name)

    def sanitize_properties(self, properties):
        """Prevent Cypher injections.

        Args:
            properties (dict): The properties to sanitize

        Returns:
            (dict): The sanitized properties
        """
        properties = copy.deepcopy(properties)
        return {self.sanitize_property_name(k): v for k, v in properties.items()}

    def wipe_database(self):
        raise NotImplementedError(
            f"{self.__class__} should implement the `wipe_database()` method"
        )

    def create_vector_index(self):
        raise NotImplementedError(
            f"{self.__class__} should implement the `create_vector_index()` method"
        )

    async def update(self, data_model, threshold=0.8):
        raise NotImplementedError(
            f"{self.__class__} should implement the `update()` method"
        )

    async def query(self, query: str, params: Dict[str, Any] = None, **kwargs):
        raise NotImplementedError(
            f"{self.__class__} should implement the `query()` method"
        )

    async def similarity_search(self, text, data_model=None, k=10, threshold=0.8):
        raise NotImplementedError(
            f"{self.__class__} should implement the `similarity_search()` method"
        )

    async def triplet_search(self, triplet_search, k=10, threshold=0.8):
        raise NotImplementedError(
            f"{self.__class__} should implement the `triplet_search()` method"
        )

    async def __repr__(self):
        return f"<DatabaseAdapter index={self.uri}>"



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from synalinks.src.api_export import synalinks_export
from synalinks.src.rewards.reward_wrappers import RewardFunctionWrapper


@synalinks_export("synalinks.rewards.exact_match")
async def exact_match(y_true, y_pred):
    """
    Computes the exact match between `y_true` and `y_pred`.

    If their values are equal, it returns a reward of 1.0; otherwise, it returns 0.0.

    Args:
        y_true (JsonDataModel): The ground truth JSON data_model.
        y_pred (JsonDataModel): The predicted JSON data_model.

    Returns:
        (float): The reward value, which is 1.0 if the values match exactly,
            and 0.0 otherwise.
    """
    reward = 0.0
    if y_pred is not None:
        if y_pred.get_json() == y_true.get_json():
            reward = 1.0
    return reward


@synalinks_export(
    [
        "synalinks.ExactMatch",
        "synalinks.rewards.ExactMatch",
    ]
)
class ExactMatch(RewardFunctionWrapper):
    """Computes the exact match between `y_true` and `y_pred`.

    Example:

    ```python
    program.compile(
        reward=synalinks.rewards.ExactMatch(),
        optimizer=synalinks.optimizers.RandomFewShot(),
    )
    ```

    Args:
        name (str): Optional. string name of the reward instance.
        in_mask (list): Optional. list of keys to keep to compute the reward.
        out_mask (list): Optional. list of keys to remove to compute the reward.
    """

    def __init__(
        self,
        name="exact_match",
        in_mask=None,
        out_mask=None,
    ):
        super().__init__(
            fn=exact_match,
            name=name,
            in_mask=in_mask,
            out_mask=out_mask,
        )

    def get_config(self):
        return {
            "name": self.name,
            "in_mask": self.in_mask,
            "out_mask": self.out_mask,
        }

    @classmethod
    def from_config(cls, config):
        return cls(**config)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from typing import Literal
from typing import Union
from unittest.mock import patch

import numpy as np

from synalinks.src import testing
from synalinks.src.backend import Entity
from synalinks.src.backend import Relation
from synalinks.src.backend import SimilaritySearch
from synalinks.src.embedding_models import EmbeddingModel
from synalinks.src.knowledge_bases.database_adapters.neo4j_adapter import Neo4JAdapter
from synalinks.src.modules import Embedding
from synalinks.src.modules import Input
from synalinks.src.programs import Program


class Document(Entity):
    label: Literal["Document"]
    text: str


class Chunk(Document):
    label: Literal["Chunk"]
    text: str


class IsPartOf(Relation):
    obj: Union[Chunk, Document]
    label: Literal["IsPartOf"]
    subj: Document


class Neo4JAdapterTest(testing.TestCase):
    @patch("litellm.aembedding")
    async def test_adapter(self, mock_embedding):
        expected_value = np.random.rand(1024)
        mock_embedding.return_value = {"data": [{"embedding": expected_value}]}

        embedding_model = EmbeddingModel(
            model="ollama/mxbai-embed-large",
        )

        adapter = Neo4JAdapter(
            uri="neo4j://localhost:7687",
            embedding_model=embedding_model,
            entity_models=[Document, Chunk],
            relation_models=[IsPartOf],
            wipe_on_start=True,
        )

        _ = await adapter.query("RETURN 1")

    @patch("litellm.aembedding")
    async def test_adapter_update_entity(self, mock_embedding):
        expected_value = np.random.rand(1024)
        mock_embedding.return_value = {"data": [{"embedding": expected_value}]}

        embedding_model = EmbeddingModel(model="ollama/mxbai-embed-large")

        adapter = Neo4JAdapter(
            uri="neo4j://localhost:7687",
            embedding_model=embedding_model,
            entity_models=[Document, Chunk],
            relation_models=[IsPartOf],
            wipe_on_start=True,
        )

        doc1 = Document(
            label="Document",
            text="test document 1",
        )

        inputs = Input(data_model=Document)
        outputs = await Embedding(
            embedding_model=embedding_model,
            in_mask=["text"],
        )(inputs)

        program = Program(
            inputs=inputs,
            outputs=outputs,
        )

        embedded_doc = await program(doc1)

        await adapter.update(embedded_doc)

    @patch("litellm.aembedding")
    async def test_adapter_update_relation(self, mock_embedding):
        expected_value = np.random.rand(1024)
        mock_embedding.return_value = {"data": [{"embedding": expected_value}]}

        embedding_model = EmbeddingModel(
            model="ollama/mxbai-embed-large",
        )

        adapter = Neo4JAdapter(
            uri="neo4j://localhost:7687",
            embedding_model=embedding_model,
            entity_models=[Document, Chunk],
            relation_models=[IsPartOf],
            wipe_on_start=True,
        )

        doc1 = Document(
            label="Document",
            text="test document 1",
        )
        doc2 = Document(
            label="Document",
            text="test document 2",
        )

        rel1 = IsPartOf(
            subj=doc1,
            label="IsPartOf",
            obj=doc2,
        )

        inputs = Input(data_model=Document)
        outputs = await Embedding(
            embedding_model=embedding_model,
            in_mask=["text"],
        )(inputs)

        program = Program(
            inputs=inputs,
            outputs=outputs,
        )

        embedded_rel = await program(rel1)

        await adapter.update(embedded_rel)

    @patch("litellm.aembedding")
    async def test_adapter_similarity_search(self, mock_embedding):
        expected_value = np.random.rand(1024)
        mock_embedding.return_value = {"data": [{"embedding": expected_value}]}

        embedding_model = EmbeddingModel(model="ollama/mxbai-embed-large")

        adapter = Neo4JAdapter(
            uri="neo4j://localhost:7687",
            embedding_model=embedding_model,
            entity_models=[Document, Chunk],
            relation_models=[IsPartOf],
            wipe_on_start=True,
        )

        inputs = Input(data_model=Document)

        outputs = await Embedding(
            embedding_model=embedding_model,
            in_mask=["text"],
        )(inputs)

        program = Program(
            inputs=inputs,
            outputs=outputs,
        )

        doc1 = Document(
            label="Document",
            text="test document 1",
        )
        doc2 = Document(
            label="Document",
            text="test document 2",
        )
        doc3 = Document(
            label="Document",
            text="test document 3",
        )

        batch = np.array([doc1, doc2, doc3], dtype=object)

        embedded_docs = await program.predict(batch)

        await adapter.update(embedded_docs[0])
        await adapter.update(embedded_docs[1])
        await adapter.update(embedded_docs[2])

        search = SimilaritySearch(
            entity_label="Document",
            similarity_search="test document",
        ).to_json_data_model()
        result = await adapter.similarity_search(search, threshold=0.0)
        self.assertTrue(len(result) > 0)

    @patch("litellm.aembedding")
    async def test_adapter_triplet_search_basic(self, mock_embedding):
        """Test basic triplet search functionality"""
        expected_value = np.random.rand(1024)
        mock_embedding.return_value = {"data": [{"embedding": expected_value}]}

        embedding_model = EmbeddingModel(model="ollama/mxbai-embed-large")

        adapter = Neo4JAdapter(
            uri="neo4j://localhost:7687",
            embedding_model=embedding_model,
            entity_models=[Document, Chunk],
            relation_models=[IsPartOf],
            wipe_on_start=True,
        )

        # Create test documents
        doc1 = Document(label="Document", text="test document 1")
        doc2 = Document(label="Document", text="test document 2")
        chunk1 = Chunk(label="Chunk", text="test chunk 1")

        # Create embeddings
        inputs = Input(data_model=Document)
        outputs = await Embedding(
            embedding_model=embedding_model,
            in_mask=["text"],
        )(inputs)
        program = Program(inputs=inputs, outputs=outputs)

        # Embed and store entities
        embedded_doc1 = await program(doc1)
        embedded_doc2 = await program(doc2)
        embedded_chunk1 = await program(chunk1)

        await adapter.update(embedded_doc1)
        await adapter.update(embedded_doc2)
        await adapter.update(embedded_chunk1)

        # Create and store relation
        rel1 = IsPartOf(subj=chunk1, label="IsPartOf", obj=doc1)
        embedded_rel = await program(rel1)
        await adapter.update(embedded_rel)

        # Test triplet search
        from synalinks.src.backend import TripletSearch

        triplet_search = TripletSearch(
            subject_label="Chunk",
            subject_similarity_search="",
            relation_label="IsPartOf",
            object_label="Document",
            object_similarity_search="",
        ).to_json_data_model()

        result = await adapter.triplet_search(triplet_search, threshold=0.0)
        self.assertTrue(len(result) > 0)
        self.assertIn("subj", result[0])
        self.assertIn("relation", result[0])
        self.assertIn("obj", result[0])

    @patch("litellm.aembedding")
    async def test_adapter_triplet_search_with_similarity(self, mock_embedding):
        """Test triplet search with similarity queries"""
        expected_value = np.random.rand(1024)
        mock_embedding.return_value = {"data": [{"embedding": expected_value}]}

        embedding_model = EmbeddingModel(model="ollama/mxbai-embed-large")

        adapter = Neo4JAdapter(
            uri="neo4j://localhost:7687",
            embedding_model=embedding_model,
            entity_models=[Document, Chunk],
            relation_models=[IsPartOf],
            wipe_on_start=True,
        )

        # Create and embed test data
        doc1 = Document(label="Document", text="artificial intelligence document")
        chunk1 = Chunk(label="Chunk", text="machine learning chapter")

        inputs = Input(data_model=Document)
        outputs = await Embedding(
            embedding_model=embedding_model,
            in_mask=["text"],
        )(inputs)
        program = Program(inputs=inputs, outputs=outputs)

        embedded_doc1 = await program(doc1)
        embedded_chunk1 = await program(chunk1)

        await adapter.update(embedded_doc1)
        await adapter.update(embedded_chunk1)

        # Create relation
        rel1 = IsPartOf(subj=chunk1, label="IsPartOf", obj=doc1)
        embedded_rel = await program(rel1)
        await adapter.update(embedded_rel)

        # Test triplet search with subject similarity
        from synalinks.src.backend import TripletSearch

        triplet_search = TripletSearch(
            subject_label="Chunk",
            subject_similarity_search="machine learning",
            relation_label="IsPartOf",
            object_label="Document",
            object_similarity_search="?",
        ).to_json_data_model()

        result = await adapter.triplet_search(triplet_search, threshold=0.0)
        self.assertTrue(len(result) > 0)

    @patch("litellm.aembedding")
    async def test_adapter_triplet_search_with_object_similarity(self, mock_embedding):
        """Test triplet search with object similarity query"""
        expected_value = np.random.rand(1024)
        mock_embedding.return_value = {"data": [{"embedding": expected_value}]}

        embedding_model = EmbeddingModel(model="ollama/mxbai-embed-large")

        adapter = Neo4JAdapter(
            uri="neo4j://localhost:7687",
            embedding_model=embedding_model,
            entity_models=[Document, Chunk],
            relation_models=[IsPartOf],
            wipe_on_start=True,
        )

        # Create and embed test data
        doc1 = Document(label="Document", text="programming tutorial")
        chunk1 = Chunk(label="Chunk", text="python basics")

        inputs = Input(data_model=Document)
        outputs = await Embedding(
            embedding_model=embedding_model,
            in_mask=["text"],
        )(inputs)
        program = Program(inputs=inputs, outputs=outputs)

        embedded_doc1 = await program(doc1)
        embedded_chunk1 = await program(chunk1)

        await adapter.update(embedded_doc1)
        await adapter.update(embedded_chunk1)

        # Create relation
        rel1 = IsPartOf(subj=chunk1, label="IsPartOf", obj=doc1)
        embedded_rel = await program(rel1)
        await adapter.update(embedded_rel)

        # Test triplet search with object similarity
        from synalinks.src.backend import TripletSearch

        triplet_search = TripletSearch(
            subject_label="Chunk",
            subject_similarity_search="?",
            relation_label="IsPartOf",
            object_label="Document",
            object_similarity_search="programming tutorial",
        ).to_json_data_model()

        result = await adapter.triplet_search(triplet_search, threshold=0.0)
        self.assertTrue(len(result) > 0)

    @patch("litellm.aembedding")
    async def test_adapter_triplet_search_with_both_similarities(self, mock_embedding):
        """Test triplet search with both subject and object similarity queries"""
        expected_value = np.random.rand(1024)
        mock_embedding.return_value = {"data": [{"embedding": expected_value}]}

        embedding_model = EmbeddingModel(model="ollama/mxbai-embed-large")

        adapter = Neo4JAdapter(
            uri="neo4j://localhost:7687",
            embedding_model=embedding_model,
            entity_models=[Document, Chunk],
            relation_models=[IsPartOf],
            wipe_on_start=True,
        )

        # Create test data
        doc1 = Document(label="Document", text="data science handbook")
        doc2 = Document(label="Document", text="machine learning guide")
        chunk1 = Chunk(label="Chunk", text="statistics chapter")
        chunk2 = Chunk(label="Chunk", text="neural networks section")

        inputs = Input(data_model=Document)
        outputs = await Embedding(
            embedding_model=embedding_model,
            in_mask=["text"],
        )(inputs)
        program = Program(inputs=inputs, outputs=outputs)

        # Embed and store all entities
        entities = [doc1, doc2, chunk1, chunk2]
        for entity in entities:
            embedded_entity = await program(entity)
            await adapter.update(embedded_entity)

        # Create relations
        rel1 = IsPartOf(subj=chunk1, label="IsPartOf", obj=doc1)
        rel2 = IsPartOf(subj=chunk2, label="IsPartOf", obj=doc2)

        for rel in [rel1, rel2]:
            embedded_rel = await program(rel)
            await adapter.update(embedded_rel)

        # Test triplet search with both similarities
        from synalinks.src.backend import TripletSearch

        triplet_search = TripletSearch(
            subject_label="Chunk",
            subject_similarity_search="statistics",
            relation_label="IsPartOf",
            object_label="Document",
            object_similarity_search="data science",
        ).to_json_data_model()

        result = await adapter.triplet_search(triplet_search, threshold=0.0)
        self.assertTrue(len(result) > 0)



from synalinks.src.knowledge_bases.database_adapters.database_adapter import (
    DatabaseAdapter,
)
from synalinks.src.knowledge_bases.database_adapters.memgraph_adapter import (
    MemGraphAdapter,
)
from synalinks.src.knowledge_bases.database_adapters.neo4j_adapter import Neo4JAdapter

# from synalinks.src.knowledge_bases.database_adapters.kuzu_adapter import KuzuAdapter


def get(uri):
    if uri.startswith("neo4j"):
        return Neo4JAdapter
    elif uri.startswith("memgraph"):
        return MemGraphAdapter
    # elif uri.startswith("kuzu"):
    #     return KuzuAdapter
    else:
        raise ValueError(f"No database adapter found for {uri}")



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from synalinks.src.datasets.arcagi import default_instructions as default_instructions
from synalinks.src.datasets.arcagi import (
    get_arcagi1_evaluation_task_names as get_arcagi1_evaluation_task_names,
)
from synalinks.src.datasets.arcagi import (
    get_arcagi1_training_task_names as get_arcagi1_training_task_names,
)
from synalinks.src.datasets.arcagi import (
    get_arcagi2_evaluation_task_names as get_arcagi2_evaluation_task_names,
)
from synalinks.src.datasets.arcagi import (
    get_arcagi2_training_task_names as get_arcagi2_training_task_names,
)
from synalinks.src.datasets.arcagi import get_input_data_model as get_input_data_model
from synalinks.src.datasets.arcagi import get_output_data_model as get_output_data_model
from synalinks.src.datasets.arcagi import load_data as load_data
from synalinks.src.datasets.arcagi import plot_task as plot_task



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

from unittest.mock import patch

from synalinks.src import testing
from synalinks.src.backend import Embeddings
from synalinks.src.embedding_models.embedding_model import EmbeddingModel


class EmbeddingModelTest(testing.TestCase):
    @patch("litellm.aembedding")
    async def test_call_api(self, mock_embedding):
        embedding_model = EmbeddingModel(model="ollama/all-minilm")

        expected_value = [0.0, 0.1, 0.2, 0.3]
        mock_embedding.return_value = {"data": [{"embedding": expected_value}]}

        result = await embedding_model(["What is the capital of France?"])
        self.assertEqual(result, Embeddings(**result).get_json())
        self.assertEqual(result, {"embeddings": [expected_value]})



# Modified from: keras/src/models/functional.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import copy
import inspect
import typing
import warnings

from synalinks.src import backend
from synalinks.src import tree
from synalinks.src.modules import Input
from synalinks.src.modules import InputModule
from synalinks.src.modules import Module
from synalinks.src.ops.function import Function
from synalinks.src.ops.function import _build_map
from synalinks.src.ops.function import make_node_key
from synalinks.src.ops.node import Node
from synalinks.src.ops.node import SynalinksHistory
from synalinks.src.programs.program import Program
from synalinks.src.saving.serialization_lib import deserialize_synalinks_object
from synalinks.src.saving.serialization_lib import serialize_synalinks_object
from synalinks.src.utils import tracking
from synalinks.src.utils.async_utils import run_maybe_nested


class Functional(Function, Program):
    """A `Functional` program is a `Program` defined as a directed graph of modules.

    Three types of `Program` exist: subclassed `Program`, `Functional` program,
    and `Sequential` (a special case of `Functional`).

    A `Functional` program can be instantiated by passing two arguments to
    `__init__()`. The first argument is the `synalinks.Input` objects
    that represent the inputs to the model.
    The second argument specifies the output data_models that represent
    the outputs of this program. Both arguments can be a nested structure
    of data_models.
    """

    def __new__(cls, *args, **kwargs):
        return typing.cast(cls, super().__new__(cls))

    @tracking.no_automatic_dependency_tracking
    def __init__(self, inputs, outputs, name=None, description=None, **kwargs):
        if isinstance(inputs, dict):
            for k, v in inputs.items():
                if isinstance(v, backend.SymbolicDataModel) and k != v.name:
                    warnings.warn(
                        "When providing `inputs` as a dict, all keys in the "
                        "dict must match the names of the corresponding "
                        f"data_models. Received key '{k}' mapping to value {v} "
                        f"which has name '{v.name}'. Change the data_model name to "
                        f"'{k}' (via `Input(..., name='{k}')`)"
                    )

        trainable = kwargs.pop("trainable", None)
        flat_inputs = tree.flatten(inputs)
        flat_outputs = tree.flatten(outputs)
        for x in flat_inputs:
            if not isinstance(x, backend.SymbolicDataModel):
                raise ValueError(
                    "All `inputs` values must be SymbolicDataModels. Received: "
                    f"inputs={inputs} including invalid value {x} of "
                    f"type {type(x)}"
                )
        for x in flat_outputs:
            if not isinstance(x, backend.SymbolicDataModel):
                raise ValueError(
                    "All `outputs` values must be SymbolicDataModels. Received: "
                    f"outputs={outputs} including invalid value {x} of "
                    f"type {type(x)}"
                )

        if not all(is_input_symbolic_data_model(t) for t in flat_inputs):
            inputs, outputs = clone_graph_nodes(inputs, outputs)

        Function.__init__(self, inputs, outputs, name=name, description=description)

        if trainable is not None:
            self.trainable = trainable

        self._modules = self.modules
        self.built = True
        self._convert_input_args = False
        self._allow_non_data_model_positional_args = True
        output_modules = [x._synalinks_history[0] for x in self.outputs]
        self.output_names = [x.name for x in output_modules]

    def _lock_state(self):
        # Unlike other modules, we allow Functional state to be mutable after
        # build. E.g. to attach a module to a model that is not part of the
        # functional DAG.
        pass

    def _obj_type(self):
        return "Functional"

    @property
    def modules(self):
        modules = []
        for operation in self._operations:
            if isinstance(operation, Module):
                modules.append(operation)
        return modules

    @modules.setter
    def modules(self, _):
        raise AttributeError(
            "`Program.modules` attribute is reserved and should not be used. "
            "Please use another name."
        )

    async def call(self, inputs, training=None):
        # Add support for training
        inputs = self._standardize_inputs(inputs)
        outputs = await self._run_through_graph(
            inputs, operation_fn=lambda op: operation_fn(op, training=training)
        )
        return unpack_singleton(outputs)

    async def compute_output_spec(self, inputs, training=None):
        # From Function
        return await super().compute_output_spec(inputs)

    async def build(self, inputs):
        self.built = True

    @property
    def input_schema(self):
        input_schemas = tree.map_structure(lambda x: x.get_schema(), self.inputs)
        if isinstance(input_schemas, (list, tuple)) and len(input_schemas) == 1:
            return input_schemas[0]
        return input_schemas

    @property
    def output_schema(self):
        output_schemas = tree.map_structure(lambda x: x.get_schema(), self.outputs)
        if isinstance(output_schemas, (list, tuple)) and len(output_schemas) == 1:
            return output_schemas[0]
        return output_schemas

    def _assert_input_compatibility(self, *args):
        return super(Program, self)._assert_input_compatibility(*args)

    def _maybe_warn_inputs_struct_mismatch(self, inputs, raise_exception=False):
        try:
            # We first normalize to tuples before performing the check to
            # suppress warnings when encountering mismatched tuples and lists.
            tree.assert_same_structure(
                tree.lists_to_tuples(inputs),
                tree.lists_to_tuples(self._inputs_struct),
            )
        except:
            model_inputs_struct = tree.map_structure(
                lambda x: x.name, self._inputs_struct
            )
            inputs_struct = tree.map_structure(lambda x: f"{x}", inputs)
            msg = (
                "The structure of `inputs` doesn't match the expected "
                f"structure.\nExpected: {model_inputs_struct}\n"
                f"Received: inputs={inputs_struct}"
            )
            if raise_exception:
                raise ValueError(msg)
            warnings.warn(msg)

    def _convert_inputs_to_json_data_models(self, flat_inputs):
        converted = []
        for x, symb_input in zip(flat_inputs, self._inputs):
            if x is None:  # TODO: check if optional
                converted.append(x)
            else:
                if backend.is_data_model(x):
                    converted.append(
                        backend.JsonDataModel(
                            value=x.json(), schema=symb_input.get_schema()
                        )
                    )
                else:
                    converted.append(x)
        return converted

    def _standardize_inputs(self, inputs):
        raise_exception = False
        if isinstance(inputs, dict) and not isinstance(self._inputs_struct, dict):
            # This is to avoid warning
            # when we have reconciable dict/list structs
            if hasattr(self._inputs_struct, "__len__") and all(
                isinstance(i, backend.SymbolicDataModel) for i in self._inputs_struct
            ):
                expected_keys = set(i.name for i in self._inputs_struct)
                keys = set(inputs.keys())
                if expected_keys.issubset(keys):
                    inputs = [inputs[i.name] for i in self._inputs_struct]
                else:
                    raise_exception = True
            elif isinstance(self._inputs_struct, backend.SymbolicDataModel):
                if self._inputs_struct.name in inputs:
                    inputs = [inputs[self._inputs_struct.name]]
                else:
                    raise_exception = True
            else:
                raise_exception = True

        self._maybe_warn_inputs_struct_mismatch(inputs, raise_exception=raise_exception)

        flat_inputs = tree.flatten(inputs)
        flat_inputs = self._convert_inputs_to_json_data_models(flat_inputs)
        return flat_inputs

    @property
    def input(self):
        return self._inputs_struct

    @property
    def output(self):
        return self._outputs_struct

    def get_config(self):
        if not functional_like_constructor(self.__class__):
            # Subclassed networks are not serializable
            # (unless serialization is implemented by
            # the author of the subclassed network).
            return Program.get_config(self)

        config = {
            "name": self.name,
            "trainable": self.trainable,
        }
        # Build a map from a module unique name (make_node_key)
        # to the index of the nodes that are saved in the config.
        # Only nodes in network_nodes are saved.
        node_reindexing_map = {}
        for operation in self.operations:
            if issubclass(operation.__class__, Functional):
                # Functional models start with a pre-existing node
                # linking their input to output.
                kept_nodes = 1
            else:
                kept_nodes = 0
            for original_node_index, node in enumerate(operation._inbound_nodes):
                node_key = make_node_key(operation, original_node_index)
                if node_key in self._nodes:
                    # i.e. we mark it to be saved
                    node_reindexing_map[node_key] = kept_nodes
                    kept_nodes += 1

        # serialize and save the modules in module_configs
        module_configs = []
        for operation in self.operations:  # From the earliest modules on.
            filtered_inbound_nodes = []
            for original_node_index, node in enumerate(operation._inbound_nodes):
                node_key = make_node_key(operation, original_node_index)
                if node_key in self._nodes:
                    # The node is relevant to the model:
                    # add to filtered_inbound_nodes.
                    node_data = serialize_node(node, own_nodes=self._nodes)
                    if node_data is not None:
                        filtered_inbound_nodes.append(node_data)

            serialize_obj_fn = serialize_synalinks_object
            module_config = serialize_obj_fn(operation)
            module_config["name"] = operation.name
            module_config["inbound_nodes"] = filtered_inbound_nodes
            module_configs.append(module_config)
        config["modules"] = module_configs

        # Gather info about inputs and outputs.
        def get_data_model_config(data_model):
            operation = data_model._synalinks_history[0]
            node_index = data_model._synalinks_history[1]
            data_model_index = data_model._synalinks_history[2]
            node_key = make_node_key(operation, node_index)
            assert node_key in self._nodes
            new_node_index = node_reindexing_map[node_key]
            return [operation.name, new_node_index, data_model_index]

        def map_data_models(data_models):
            if isinstance(data_models, backend.SymbolicDataModel):
                return [get_data_model_config(data_models)]
            return tree.map_structure(get_data_model_config, data_models)

        config["input_modules"] = map_data_models(self._inputs_struct)
        config["output_modules"] = map_data_models(self._outputs_struct)
        return copy.deepcopy(config)


def functional_from_config(cls, config, custom_objects=None):
    """Instantiates a Functional program from its config (from `get_config()`).

    Args:
        cls: Class of the program, e.g. a custom subclass of `Program`.
        config: Output of `get_config()` for the original model instance.
        custom_objects: Optional dict of custom objects.

    Returns:
        An instance of `cls`.
    """
    # Module instances created during
    # the graph reconstruction process
    created_modules = {}

    # Dictionary mapping module instances to
    # node data that specifies a module call.
    # It acts as a queue that maintains any unprocessed
    # module call until it becomes possible to process it
    # (i.e. until the input data models to the call all exist).
    unprocessed_nodes = {}

    def add_unprocessed_node(module, node_data):
        """Add node to module list

        Arg:
            module: module object
            node_data: Node data specifying module call
        """
        if module not in unprocessed_nodes:
            unprocessed_nodes[module] = [node_data]
        else:
            unprocessed_nodes[module].append(node_data)

    def process_node(module, node_data):
        """Reconstruct node by linking to inbound modules

        Args:
            module: Module to process
            node_data: List of module configs
        """
        args, kwargs = deserialize_node(node_data, created_modules)
        # Call module on its inputs, thus creating the node
        # and building the module if needed.
        run_maybe_nested(module(*args, **kwargs))

    def process_module(module_data):
        """Deserializes a module and index its inbound nodes.

        Args:
            module_data: module config dict.
        """
        module_name = module_data["name"]

        # Instantiate module.
        module = deserialize_synalinks_object(module_data, custom_objects=custom_objects)
        created_modules[module_name] = module

        # Gather module inputs.
        inbound_nodes_data = module_data["inbound_nodes"]
        for node_data in inbound_nodes_data:
            # We don't process nodes (i.e. make module calls)
            # on the fly because the inbound node may not yet exist,
            # in case of module shared at different topological depths
            # (e.g. a model such as A(B(A(B(x)))))
            add_unprocessed_node(module, node_data)

    # Extract config used to instantiate Functional model from the config. The
    # remaining config will be passed as keyword arguments to the Model
    # constructor.
    functional_config = {}
    for key in ["modules", "input_modules", "output_modules"]:
        functional_config[key] = config.pop(key)
    for key in ["name", "trainable"]:
        if key in config:
            functional_config[key] = config.pop(key)
        else:
            functional_config[key] = None

    # First, we create all modules and enqueue nodes to be processed
    for module_data in functional_config["modules"]:
        process_module(module_data)

    # Then we process nodes in order of module depth.
    # Nodes that cannot yet be processed (if the inbound node
    # does not yet exist) are re-enqueued, and the process
    # is repeated until all nodes are processed.
    while unprocessed_nodes:
        for module_data in functional_config["modules"]:
            module = created_modules[module_data["name"]]

            # Process all nodes in module, if not yet processed
            if module in unprocessed_nodes:
                node_data_list = unprocessed_nodes[module]

                # Process nodes in order
                node_index = 0
                while node_index < len(node_data_list):
                    node_data = node_data_list[node_index]
                    try:
                        process_node(module, node_data)

                    # If the node does not have all inbound modules
                    # available, stop processing and continue later
                    except IndexError:
                        break

                    node_index += 1

                # If not all nodes processed then store unprocessed nodes
                if node_index < len(node_data_list):
                    unprocessed_nodes[module] = node_data_list[node_index:]
                # If all nodes processed remove the module
                else:
                    del unprocessed_nodes[module]

    # Create list of input and output data models and return new class
    name = functional_config["name"]
    trainable = functional_config["trainable"]

    def get_data_model(module_name, node_index, data_model_index):
        assert module_name in created_modules
        module = created_modules[module_name]
        if isinstance(module, Functional):
            # Functional models start out with a built-in node.
            node_index -= 1
        module_output_data_models = module._inbound_nodes[node_index].output_data_models
        return module_output_data_models[data_model_index]

    def map_data_models(data_models):
        if (
            isinstance(data_models, list)
            and len(data_models) == 3
            and isinstance(data_models[0], str)
        ):
            # Leaf
            return get_data_model(*data_models)
        if isinstance(data_models, dict):
            return {k: map_data_models(v) for k, v in data_models.items()}
        if isinstance(data_models, tuple):
            return tuple([map_data_models(v) for v in data_models])
        return [map_data_models(v) for v in data_models]

    input_data_models = map_data_models(functional_config["input_modules"])
    output_data_models = map_data_models(functional_config["output_modules"])
    if isinstance(input_data_models, list) and len(input_data_models) == 1:
        input_data_models = input_data_models[0]
    if isinstance(output_data_models, list) and len(output_data_models) == 1:
        output_data_models = output_data_models[0]

    return cls(
        inputs=input_data_models,
        outputs=output_data_models,
        name=name,
        trainable=trainable,
        **config,
    )


def operation_fn(operation, training):
    def call(*args, **kwargs):
        if (
            hasattr(operation, "_call_has_training_arg")
            and operation._call_has_training_arg
            and training is not None
        ):
            kwargs["training"] = training
        return operation(*args, **kwargs)

    return call


def functional_like_constructor(cls):
    init_args = inspect.getfullargspec(cls.__init__).args[1:]
    functional_init_args = inspect.getfullargspec(Functional.__init__).args[1:]
    if init_args == functional_init_args:
        return True
    return False


def unpack_singleton(x):
    if isinstance(x, (list, tuple)) and len(x) == 1:
        return x[0]
    return x


def serialize_node(node, own_nodes=()):
    if not node.input_data_models:
        # Does not need to be serialized.
        return

    def serialize_symbolic_datamodel(x):
        # Serialize SymbolicDataModel while converting
        # node indices to only include nodes relevant to `own_nodes`.
        if isinstance(x, backend.SymbolicDataModel):
            operation, node_index, data_model_index = x._synalinks_history
            irrelevant_node_count = 0
            for i, node in enumerate(operation._inbound_nodes[:node_index]):
                node_key = make_node_key(operation, i)
                if node_key not in own_nodes:
                    irrelevant_node_count += 1
            x._synalinks_history = SynalinksHistory(
                operation, node_index - irrelevant_node_count, data_model_index
            )
            serialized = serialize_synalinks_object(x)
            x._synalinks_history = SynalinksHistory(
                operation, node_index, data_model_index
            )
            return serialized
        return x

    args = node.arguments.args
    kwargs = node.arguments.kwargs

    args = tree.map_structure(serialize_symbolic_datamodel, args)
    kwargs = tree.map_structure(serialize_symbolic_datamodel, kwargs)
    return {
        "args": serialize_synalinks_object(args),
        "kwargs": serialize_synalinks_object(kwargs),
    }


def deserialize_node(node_data, created_modules):
    """Return (args, kwargs) for calling the node module."""
    if not node_data:
        return [], {}

    if isinstance(node_data, list):
        # Legacy case.
        input_data_models = []
        for input_data in node_data:
            inbound_module_name = input_data[0]
            inbound_node_index = input_data[1]
            inbound_data_model_index = input_data[2]
            if len(input_data) == 3:
                kwargs = {}
            elif len(input_data) == 4:
                kwargs = input_data[3]
            else:
                raise ValueError("Cannot deserialize the program (invalid config data?)")
            inbound_module = created_modules[inbound_module_name]

            # Raise an error if the corresponding module node
            # has not yet been created
            if len(inbound_module._inbound_nodes) <= inbound_node_index:
                raise IndexError(
                    "Module node index out of bounds.\n"
                    f"inbound_module = {inbound_module}\n"
                    "inbound_module._inbound_nodes = "
                    f"{inbound_module._inbound_nodes}\n"
                    f"inbound_node_index = {inbound_node_index}"
                )
            inbound_node = inbound_module._inbound_nodes[inbound_node_index]
            input_data_models.append(
                inbound_node.output_data_models[inbound_data_model_index]
            )
        return [unpack_singleton(input_data_models)], kwargs

    args = deserialize_synalinks_object(node_data["args"])
    kwargs = deserialize_synalinks_object(node_data["kwargs"])

    def convert_revived_data_model(x):
        if isinstance(x, backend.SymbolicDataModel):
            history = x._pre_serialization_synalinks_history
            if history is None:
                return x
            module = created_modules.get(history[0], None)
            if module is None:
                raise ValueError(f"Unknown module: {history[0]}")
            inbound_node_index = history[1]
            inbound_data_model_index = history[2]
            if len(module._inbound_nodes) <= inbound_node_index:
                raise IndexError(
                    "Module node index out of bounds.\n"
                    f"inbound_module = {module}\n"
                    f"inbound_module._inbound_nodes = {module._inbound_nodes}\n"
                    f"inbound_node_index = {inbound_node_index}"
                )
            inbound_node = module._inbound_nodes[inbound_node_index]
            return inbound_node.output_data_models[inbound_data_model_index]
        return x

    args = tree.map_structure(convert_revived_data_model, args)
    kwargs = tree.map_structure(convert_revived_data_model, kwargs)
    return args, kwargs


def is_input_symbolic_data_model(x):
    (
        operation,
        node_index,
        _,
    ) = x._synalinks_history
    node = operation._inbound_nodes[node_index]
    return node.is_input


def clone_single_symbolic_data_model(x):
    return backend.SymbolicDataModel(schema=x.get_schema(), name=x.name + "_clone")


def clone_symbolic_data_models(data_models, sd_id_mapping):
    def swap(x):
        if not isinstance(x, backend.SymbolicDataModel):
            return x
        if id(x) in sd_id_mapping:
            return sd_id_mapping[id(x)]
        new_x = clone_single_symbolic_data_model(x)
        sd_id_mapping[id(x)] = new_x
        return new_x

    return tree.map_structure(swap, data_models)


def find_nodes_by_inputs_and_outputs(inputs, outputs):
    nodes, _ = _build_map(inputs, outputs)
    return nodes


def clone_graph_nodes(inputs, outputs):
    """Clone the `Node` between the inputs and output data_models.

    This function is used to create a new functional program from any intermediate
    symbolic data models. The clone of the nodes mimic the behavior of reconstructing
    the functional graph network by re-executing all the `__call__()` methods.
    The cloned nodes will be appended to the modules.

    Note that a new `synalinks.Input` will be created for any items in the
    `inputs`

    Args:
    inputs: A nested structure of `SymbolicDataModel` instances.
    outputs: A nested structure of `SymbolicDataModel` instances.

    Returns:
        A pair of inputs and outputs, with cloned `SymbolicDataModel` instances.
        They can be used to create a new functional program.
    """
    nodes_to_clone = find_nodes_by_inputs_and_outputs(inputs, outputs)
    cloned_inputs = []
    cloned_outputs = []
    # We not only need to create copies of Nodes (mimic the calls), also need to
    # clone symbolic data models to avoid the override of _synalinks_history attached on
    # the symbolic data model. The following dict is used to track any synalinks
    # data_model we cloned. The key is the string ID of the original synalinks
    # data_model, and value is the cloned symbolic data model instance.
    sd_id_mapping = {}
    op_id_mapping = {}

    for sd_input in tree.flatten(inputs):
        if is_input_symbolic_data_model(sd_input):
            # For any existing symbolic data model from synalinks.Input, leave them as is.
            cloned_inputs.append(sd_input)
            sd_id_mapping[id(sd_input)] = sd_input
        else:
            # We need to create a new symbolic data model for any intermediate data_model
            cloned_input = Input(
                schema=sd_input.get_schema(),
                name=sd_input.name + "CLONE",
            )
            cloned_inputs.append(cloned_input)
            sd_id_mapping[id(sd_input)] = cloned_input
            op_id_mapping[id(sd_input._synalinks_history[0])] = (
                cloned_input._synalinks_history[0]
            )
    cloned_inputs = tree.pack_sequence_as(inputs, cloned_inputs)

    for sd_output in tree.flatten(outputs):
        cpy = clone_single_symbolic_data_model(sd_output)
        # We reuse the _synalinks_history here, which contains the old information.
        cpy._synalinks_history = sd_output._synalinks_history
        cloned_outputs.append(cpy)
        sd_id_mapping[id(sd_output)] = cpy
    cloned_outputs = tree.pack_sequence_as(outputs, cloned_outputs)

    for node in nodes_to_clone:
        if id(node.operation) in op_id_mapping:
            operation = op_id_mapping[id(node.operation)]
        else:
            operation = node.operation
        # Clone any symbolic data model to avoid override of _synalinks_history
        # Or reuse an existing symbolic data model if it has already been cloned.
        output_copy = clone_symbolic_data_models(node.output_data_models, sd_id_mapping)
        if not isinstance(operation, InputModule):
            call_args_copy = clone_symbolic_data_models(
                node.arguments.args, sd_id_mapping
            )
            call_kwargs_copy = clone_symbolic_data_models(
                node.arguments.kwargs, sd_id_mapping
            )
        else:
            call_args_copy = ()
            call_kwargs_copy = {}
        # Creating new nodes based on the existing node information.  Node wires
        # itself to inbound and outbound modules.  The Node constructor actually
        # updates this module's self._inbound_nodes, sets _synalinks_history on the
        # outputs, and adds itself to the `_outbound_nodes` of the modules that
        # produced the inputs to this module call.
        Node(
            operation,
            call_args=call_args_copy,
            call_kwargs=call_kwargs_copy,
            outputs=output_copy,
        )
    return cloned_inputs, cloned_outputs



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import json
import logging

from synalinks.src import tree
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import any_symbolic_data_models
from synalinks.src.hooks.hook import Hook

_SYMBOLIC_LOG_TEMPLATE = """
\033[94m---
# {name}
Call ID: {call_id}
Module Name: {module_name}
Module Description: {module_description}
Data Model JSON Schema:
{data_model_schema}
---\033[0m
"""

_DATA_LOG_TEMPLATE = """
\033[92m---
# {name}
Call ID: {call_id}
Module Name: {module_name}
Module Description: {module_description}
Data Model JSON:
{data_model_json}
---\033[0m
"""

_EXCEPTION_TEMPLATE = """
\033[91m---
# Exception
Call ID: {call_id}
Module Name: {module_name}
Module Description: {module_description}
Exception: {exception}
---\033[0m
"""


@synalinks_export("synalinks.hooks.Logger")
class Logger(Hook):
    """Logger hook for logging module calls.

    This hook is set by default when you enables logging.

    Example:

    ```python
    import synalinks

    synalinks.enable_logging()
    ```
    """

    def _maybe_setup_logger(self):
        if not hasattr(self, "logger"):
            self.logger = logging.getLogger(self.module.name)

    def on_call_begin(
        self,
        call_id,
        inputs=None,
    ):
        self._maybe_setup_logger()
        if not inputs:
            return
        module_name = self.module.name
        module_description = self.module.description
        flatten_inputs = tree.flatten(inputs)
        if any_symbolic_data_models(inputs):
            data_models_schemas = [
                dm.get_schema() for dm in flatten_inputs if dm is not None
            ]
            if data_models_schemas:
                self.logger.info(
                    _SYMBOLIC_LOG_TEMPLATE.format(
                        name="Symbolic Call Start",
                        call_id=call_id,
                        module_name=module_name,
                        module_description=module_description,
                        data_model_schema=json.dumps(
                            data_models_schemas,
                            indent=2,
                        ),
                    ),
                )
        else:
            data_models_jsons = [dm.get_json() for dm in flatten_inputs if dm is not None]
            if data_models_jsons:
                self.logger.info(
                    _DATA_LOG_TEMPLATE.format(
                        name="Call Start",
                        call_id=call_id,
                        module_name=module_name,
                        module_description=module_description,
                        data_model_json=json.dumps(
                            data_models_jsons,
                            indent=2,
                        ),
                    )
                )

    def on_call_end(
        self,
        call_id,
        outputs=None,
        exception=None,
    ):
        self._maybe_setup_logger()
        module_name = self.module.name
        module_description = self.module.description
        if exception:
            self.logger.error(
                _EXCEPTION_TEMPLATE.format(
                    call_id=call_id,
                    exception=exception,
                    module_name=module_name,
                    module_description=module_description,
                )
            )
        if not outputs:
            return
        flatten_outputs = tree.flatten(outputs)
        if any_symbolic_data_models(outputs):
            data_models_schemas = [
                dm.get_schema() for dm in flatten_outputs if dm is not None
            ]
            if data_models_schemas:
                self.logger.info(
                    _SYMBOLIC_LOG_TEMPLATE.format(
                        name="Symbolic Call End",
                        call_id=call_id,
                        module_name=module_name,
                        module_description=module_description,
                        data_model_schema=json.dumps(
                            data_models_schemas,
                            indent=2,
                        ),
                    ),
                )
        else:
            data_models_jsons = [
                dm.get_json() for dm in flatten_outputs if dm is not None
            ]
            if data_models_jsons:
                self.logger.info(
                    _DATA_LOG_TEMPLATE.format(
                        name="Call End",
                        call_id=call_id,
                        module_name=module_name,
                        module_description=module_description,
                        data_model_json=json.dumps(
                            data_models_jsons,
                            indent=2,
                        ),
                    )
                )



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import inspect

import pydantic
from typing_extensions import ClassVar

from synalinks.src import tree
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend.common.json_data_model import JsonDataModel
from synalinks.src.backend.common.symbolic_data_model import SymbolicDataModel
from synalinks.src.saving.synalinks_saveable import SynalinksSaveable
from synalinks.src.utils.async_utils import run_maybe_nested

IS_THREAD_SAFE = True


class MetaDataModel(type(pydantic.BaseModel)):
    """The metaclass data model.

    This class defines operations at the metaclass level.
    Allowing to use Synalinks Python operators with `DataModel` types.
    """

    def __add__(cls, other):
        """Concatenates this data model with another.

        Args:
            other (SymbolicDataModel | DataModel):
                The other data model to concatenate with.

        Returns:
            (SymbolicDataModel): The concatenated data model.
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.Concat().symbolic_call(cls, other))

    def __radd__(cls, other):
        """Concatenates (reverse) another data model with this one.

        Args:
            other (SymbolicDataModel | DataModel):
                The other data model to concatenate with.

        Returns:
            (SymbolicDataModel): The concatenated data model.
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.Concat().symbolic_call(other, cls))

    def __and__(cls, other):
        """Perform a `logical_and` with another data model.

        If one of them is `None`, output `None`. If both are provided,
        then concatenates this data model with the other.

        Args:
            other (SymbolicDataModel | DataModel): The other data model to concatenate
                with.

        Returns:
            (SymbolicDataModel | None): The concatenated data model or `None`
                based on the `logical_and` table.
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.And().symbolic_call(cls, other))

    def __rand__(cls, other):
        """Perform a `logical_and` (reverse) with another data model.

        If one of them is `None`, output `None`. If both are provided,
        then concatenates the other data model with this one.

        Args:
            other (SymbolicDataModel | DataModel): The other data model to concatenate
                with.

        Returns:
            (SymbolicDataModel | None): The concatenated data model or `None`
                based on the `logical_and` table.
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.And().symbolic_call(other, cls))

    def __or__(cls, other):
        """Perform a `logical_or` with another data model.

        If one of them is `None`, output the other one. If both are provided,
        then concatenates this data model with the other.

        Args:
            other (SymbolicDataModel): The other data model to concatenate with.

        Returns:
            (SymbolicDataModel | None): The concatenation of data model if both are
                provided, or the non-None data model or None if none are provided.
                (See `logical_or` table).
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.Or().symbolic_call(cls, other))

    def __ror__(cls, other):
        """Perform a `logical_or` (reverse) with another data model.

        If one of them is `None`, output the other one. If both are provided,
        then concatenates the other data model with this one.

        Args:
            other (SymbolicDataModel | DataModel): The other data model to concatenate
                with.

        Returns:
            (SymbolicDataModel | None): The concatenation of data model if both are
                provided, or the non-None data model or None if none are provided.
                (See `logical_or` table).
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.Or().symbolic_call(other, cls))

    def __xor__(cls, other):
        """Perform a `logical_xor` with another data model.

        If one of them is `None`, output the other one. If both are provided,
        then the output is `None`.

        Args:
            other (SymbolicDataModel): The other data model to concatenate with.

        Returns:
            (SymbolicDataModel | None): `None` if both are
                provided, or the non-None data model if one is provided
                or `None` if none are provided. (See `logical_xor` table).
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.Xor().symbolic_call(cls, other))

    def __rxor__(cls, other):
        """Perform a `logical_xor` (reverse) with another data model.

        If one of them is None, output the other one. If both are provided,
        then concatenates the other data model with this one.

        Args:
            other (SymbolicDataModel | DataModel): The other data model to concatenate
                with.

        Returns:
            (SymbolicDataModel | None): `None` if both are
                provided, or the non-None data model if one is provided
                or `None` if none are provided. (See `logical_xor` table).
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.Xor().symbolic_call(other, cls))

    def __contains__(cls, other):
        """Check if the schema of `other` is contained in this one.

        Args:
            other (SymbolicDataModel | DataModel): The other data model to compare with.

        Returns:
            (bool): True if all properties of `other` are present in this one.
        """
        from synalinks.src.backend.common.json_schema_utils import contains_schema

        return contains_schema(cls.get_schema(), other.get_schema())

    def factorize(cls):
        """Factorizes the data model.

        Returns:
            (SymbolicDataModel): The factorized data model.
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.Factorize().symbolic_call(cls))

    def in_mask(cls, mask=None, recursive=True):
        """Applies a mask to **keep only** specified keys of the data model.

        Args:
            mask (list): The mask to be applied (list of keys).
            recursive (bool): Optional. Whether to apply the mask recursively.
                Defaults to `True`.

        Returns:
            (SymbolicDataModel): The data model with the mask applied.
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.InMask(mask=mask, recursive=True).symbolic_call(cls))

    def out_mask(cls, mask=None, recursive=True):
        """Applies an mask to **remove** specified keys of the data model.

        Args:
            mask (list): The mask to be applied (list of keys).
            recursive (bool): Optional. Whether to apply the mask recursively.
                Defaults to `True`.

        Returns:
            (SymbolicDataModel): The data model with the mask applied.
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.OutMask(mask=mask, recursive=True).symbolic_call(cls))

    def prefix(cls, prefix=None):
        """Add a prefix to **all** the data model fields (non-recursive).

        Args:
            prefix (str): the prefix to add

        Returns:
            (SymbolicDataModel): The data model with the prefix added.
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.Prefix(prefix=prefix).symbolic_call(cls))

    def suffix(cls, suffix=None):
        """Add a suffix to **all** the data model fields (non-recursive).

        Args:
            suffix (str): the suffix to add

        Returns:
            (SymbolicDataModel): The data model with the suffix added.
        """
        from synalinks.src import ops

        return run_maybe_nested(ops.Suffix(suffix=suffix).symbolic_call(cls))


class DataModel(pydantic.BaseModel, SynalinksSaveable, metaclass=MetaDataModel):
    """The backend-dependent data model.

    This data model uses Pydantic to provide, JSON schema inference
    and JSON serialization.

    Examples:

    **Creating a DataModel for structured output**

    ```python
    class AnswerWithReflection(synalinks.DataModel):
        thinking: str = synalinks.Field(
            description="Your step by step thinking",
        )
        reflection: str = synalinks.Field(
            description="The reflection about your thinking",
        )
        answer: str = synalinks.Field(
            description="The correct answer",
        )

    language_model = synalinks.LanguageModel("ollama/mistral")

    generator = synalinks.Generator(
        data_model=AnswerWithReflection,
        language_model=language_model,
    )
    ```
    """

    model_config: ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="forbid")

    @classmethod
    def get_schema(cls):
        """Gets the JSON schema of the data model.

        Returns:
            (dict): The JSON schema.
        """
        return cls.model_json_schema()

    @classmethod
    def keys(cls):
        """Gets the JSON properties keys of the data model.

        Returns:
            (dict): The JSON schema.
        """
        schema = cls.get_schema()
        if "properties" in schema:
            return schema["properties"].keys()
        else:
            return []

    @classmethod
    def prettify_schema(cls):
        """Get a pretty version of the JSON schema for display.

        Returns:
            (str): The indented JSON schema.
        """
        import json

        return json.dumps(cls.get_schema(), indent=2)

    @classmethod
    def to_symbolic_data_model(cls, name=None):
        """Converts the data model to a symbolic data model.

        Args:
            name (str): Optional. The name of the symbolic data model.
                If None, a name will be given automatically.

        Returns:
            (SymbolicDataModel): The symbolic data model.
        """
        return SymbolicDataModel(schema=cls.get_schema(), name=name)

    def get_json(self):
        """Gets the JSON value of the data model.

        Returns:
            (dict): The JSON value.
        """
        return self.model_dump(mode="json")

    def prettify_json(self):
        """Get a pretty version of the JSON object for display.

        Returns:
            (str): The indented JSON object.
        """
        import json

        return json.dumps(self.get_json(), indent=2)

    def __repr__(self):
        return f"<DataModel json={self.get_json()}, schema={self.get_schema()}>"

    def get(self, key, default=None):
        """Get wrapper to make it easier to access JSON fields.

        Args:
            key (str): The key to access.
            default (any): The default value if key not found.
        """
        try:
            return self.__getattribute__(key)
        except Exception:
            return default

    def __getitem__(self, key):
        """Get item wrapper to make it easier to access JSON fields.

        Args:
            key (str): The key to access.
        """
        return self.__getattribute__(key)

    def to_json_data_model(self, name=None):
        """Converts the data model to a backend-independent data model.

        Args:
            name (str): Optional. The name of the json data model.
                If None, a name will be given automatically.

        Returns:
            (JsonDataModel): The backend-independent data model.
        """
        return JsonDataModel(
            schema=self.get_schema(),
            json=self.get_json(),
            name=name,
        )

    def __add__(self, other):
        """Concatenates this data model with another.

        Args:
            other (JsonDataModel | DataModel | SymbolicDataModel):
                The other data model to concatenate with.

        Returns:
            (JsonDataModel | SymbolicDataModel): The concatenated data model.
                If one of them is a metaclass or symbolic data model,
                then output a `SymbolicDataModel`.
        """
        from synalinks.src import ops

        if any_meta_class(self, other):
            return run_maybe_nested(ops.Concat().symbolic_call(self, other))
        else:
            return run_maybe_nested(ops.Concat()(self, other))

    def __radd__(self, other):
        """Concatenates another data model with this one.

        Args:
            other (JsonDataModel | DataModel | SymbolicDataModel):
                The other data model to concatenate with.

        Returns:
            (JsonDataModel | SymbolicDataModel): The concatenated data model.
                If one of them is a metaclass or symbolic data model,
                then output a `SymbolicDataModel`.
        """
        from synalinks.src import ops

        if any_meta_class(self, other):
            return run_maybe_nested(
                ops.Concat().symbolic_call(other, self),
            )
        else:
            return run_maybe_nested(
                ops.Concat()(other, self),
            )

    def __and__(self, other):
        """Perform a `logical_and` with another data model.

        If one of them is None, output None. If both are provided,
        then concatenates the other data model with this one.

        If the other is a metaclass or symbolic data model, output a symbolic data model.

        Args:
            other (JsonDataModel | SymbolicDataModel | DataModel):
                The other data model to concatenate with.

        Returns:
            (JsonDataModel | SymbolicDataModel | None): The concatenated data model or
                `None` based on the `logical_and` table.
        """
        from synalinks.src import ops

        if any_meta_class(self, other):
            return run_maybe_nested(
                ops.Add().symbolic_call(self, other),
            )
        else:
            return run_maybe_nested(
                ops.Add()(self, other),
            )

    def __rand__(self, other):
        """Perform a `logical_and` (reverse) with another data model.

        If one of them is None, output None. If both are provided,
        then concatenates the other data model with this one.

        If the other is a metaclass or symbolic data model, output a symbolic data model.

        Args:
            other (JsonDataModel | SymbolicDataModel | DataModel):
                The other data model to concatenate with.

        Returns:
            (JsonDataModel | SymbolicDataModel | None): The concatenated data model or
                `None` based on the `logical_and` table.
        """
        from synalinks.src import ops

        if any_meta_class(other, self):
            return run_maybe_nested(
                ops.Add().symbolic_call(other, self),
            )
        else:
            return run_maybe_nested(
                ops.Add()(other, self),
            )

    def __or__(self, other):
        """Perform a `logical_or` with another data model

        If one of them is None, output the other one. If both are provided,
        then concatenates this data model with the other.

        If the other is a metaclass or symbolic data model, output a symbolic data model.

        Args:
            other (JsonDataModel | SymbolicDataModel | DataModel):
                The other data model to concatenate with.

        Returns:
            (JsonDataModel | SymbolicDataModel | None): The concatenation of data model
                if both are provided, or the non-None data model or None if none are
                provided. (See `logical_or` table).
        """
        from synalinks.src import ops

        if any_meta_class(self, other):
            return run_maybe_nested(
                ops.Or().symbolic_call(self, other),
            )
        else:
            return run_maybe_nested(
                ops.Or()(self, other),
            )

    def __ror__(self, other):
        """Perform a `logical_or` (reverse) with another data model

        If one of them is None, output the other one. If both are provided,
        then concatenates the other data model with this one.

        If the other is a metaclass or symbolic data model, output a symbolic data model.

        Args:
            other (JsonDataModel | SymbolicDataModel | DataModel):
                The other data model to concatenate with.

        Returns:
            (JsonDataModel | SymbolicDataModel | None): The concatenation of data model
                if both are provided, or the non-None data model or None if none are
                provided. (See `logical_or` table).
        """
        from synalinks.src import ops

        if any_meta_class(other, self):
            return run_maybe_nested(
                ops.Or().symbolic_call(other, self),
            )
        else:
            return run_maybe_nested(
                ops.Or()(other, self),
            )

    def __xor__(self, other):
        """Perform a `logical_xor` with another data model.

        If one of them is None, output the other one. If both are provided,
        then output None.

        Args:
            other (SymbolicDataModel): The other data model to concatenate with.

        Returns:
            (SymbolicDataModel | None): `None` if both are
                provided, or the non-None data model if one is provided
                or `None` if none are provided. (See `logical_xor` table).
        """
        from synalinks.src import ops

        if any_meta_class(self, other):
            return run_maybe_nested(
                ops.Xor().symbolic_call(self, other),
            )
        else:
            return run_maybe_nested(
                ops.Xor()(self, other),
            )

    def __rxor__(self, other):
        """Perform a `logical_xor` (reverse) with another data model.

        If one of them is None, output the other one. If both are provided,
        then output None.

        Args:
            other (SymbolicDataModel | DataModel): The other data model to concatenate
                with.

        Returns:
            (SymbolicDataModel | None): `None` if both are
                provided, or the non-None data model if one is provided
                or `None` if none are provided. (See `logical_xor` table).
        """
        from synalinks.src import ops

        if any_meta_class(other, self):
            return run_maybe_nested(
                ops.Xor().symbolic_call(other, self),
            )
        else:
            return run_maybe_nested(
                ops.Xor()(other, self),
            )

    def __contains__(cls, other):
        """Check if the schema of `other` is contained in this one.

        Args:
            other (SymbolicDataModel | DataModel): The other data model to compare with.

        Returns:
            (bool): True if all properties of `other` are present in this one.
        """
        from synalinks.src.backend.common.json_schema_utils import contains_schema

        return contains_schema(cls.get_schema(), other.get_schema())

    def get_config(self):
        return self.get_json()

    @classmethod
    def from_config(cls, config):
        return cls(**config)


def is_data_model(x):
    """Returns whether `x` is a DataModel.

    Args:
        x (any): The object to check.

    Returns:
        (bool): True if `x` is a DataModel, False otherwise.
    """
    return isinstance(x, DataModel)


def any_data_model(args=None, kwargs=None):
    """Check if any of the arguments are backend-dependent data models.

    Args:
        args (tuple): Optional. The positional arguments to check.
        kwargs (dict): Optional. The keyword arguments to check.

    Returns:
        (bool): True if any of the arguments are meta classes, False otherwise.
    """
    args = args or ()
    kwargs = kwargs or {}
    for x in tree.flatten((args, kwargs)):
        if is_meta_class(x):
            return True
    return False


def any_meta_class(args=None, kwargs=None):
    """Check if any of the arguments are meta classes.

    This happen when using a `DataModel` without instanciating it.
    In Synalinks this is used when declaring data models for schema inference.

    Args:
        args (tuple): Optional. The positional arguments to check.
        kwargs (dict): Optional. The keyword arguments to check.

    Returns:
        (bool): True if any of the arguments are meta classes, False otherwise.
    """
    args = args or ()
    kwargs = kwargs or {}
    for x in tree.flatten((args, kwargs)):
        if is_meta_class(x):
            return True
    return False


@synalinks_export(
    [
        "synalinks.utils.is_meta_class",
        "synalinks.backend.is_meta_class",
    ]
)
def is_meta_class(x):
    """Returns whether `x` is a meta class.

    A meta class is a python type. This method checks if the data model provided
    if a meta class, allowing to detect if the `DataModel` have been instanciated.
    Meta classes are using in Synalinks when declaring data models for schema inference.

    Args:
        x (any): The object to check.

    Returns:
        (bool): True if `x` is a meta class, False otherwise.
    """
    return inspect.isclass(x)



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import asyncio
import uuid

from synalinks.src import ops
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import ChatMessage
from synalinks.src.backend import ChatMessages
from synalinks.src.backend import ChatRole
from synalinks.src.backend import JsonDataModel
from synalinks.src.backend import SymbolicDataModel
from synalinks.src.backend import ToolCall
from synalinks.src.backend import is_chat_messages
from synalinks.src.backend.common.dynamic_json_schema_utils import dynamic_tool_calls
from synalinks.src.backend.common.json_utils import out_mask_json
from synalinks.src.modules.module import Module
from synalinks.src.modules.ttc.chain_of_thought import ChainOfThought
from synalinks.src.saving import serialization_lib


def get_default_instructions():
    """The default parallel function calling agent instructions."""
    return """
Think step by step: Use the thinking field to elaborate what you observe and what do you need to accomplish next.
Reflect on prior steps: Review your previous actions and their outcomes to avoid unnecessary repetition.
Avoid unnecessary actions: If you already have enough information to complete the user task, return an empty tool calls array.
""".strip()


@synalinks_export(
    [
        "synalinks.modules.FunctionCallingAgent",
        "synalinks.FunctionCallingAgent",
    ]
)
class FunctionCallingAgent(Module):
    """A trainable parallel function calling agent.

    The agent has 2 different modes:

    - Autonomous: It will execute tools as soon as called.
    - Non-autonomous: It will return the tool arguments as a ChatMessage.

    In *autonomous* mode, the agent accept **any kind of data model input** and perform a final inference to
    eventually format its final answer if a `data_model` or `schema` is provided.

    Example:

    ```python
    import synalinks
    import asyncio

    class Query(synalinks.DataModel):
        query: str = synalinks.Field(
            description="The user query",
        )

    class NumericalFinalAnswer(synalinks.DataModel):
        final_answer: float = synalinks.Field(
            description="The correct final numerical answer",
        )

    async def calculate(expression: str):
        \"""Calculate the result of a mathematical expression.

        Args:
            expression (str): The mathematical expression to calculate, such as
                '2 + 2'. The expression can contain numbers, operators (+, -, *, /),
                parentheses, and spaces.
        \"""
        if not all(char in "0123456789+-*/(). " for char in expression):
            return {
                "result": None,
                "log": (
                        "Error: invalid characters in expression. "
                        "The expression can only contain numbers, operators (+, -, *, /),"
                        " parentheses, and spaces NOT letters."
                    ),
            }
        try:
            # Evaluate the mathematical expression safely
            result = round(float(eval(expression, {"__builtins__": None}, {})), 2)
            return {
                "result": result,
                "log": "Successfully executed",
            }
        except Exception as e:
            return {
                "result": None,
                "log": f"Error: {e}",
            }

    async def main():
        language_model = synalinks.LanguageModel(model="ollama/mistral")

        tools = [
            synalinks.Tool(calculate),
        ]

        inputs = synalinks.Input(data_model=Query)
        outputs = await synalinks.FunctionCallingAgent(
            data_model=NumericalFinalAnswer,
            tools=tools,
            language_model=language_model,
            max_iterations=5,
            autonomous=True,
        )(inputs)
        agent = synalinks.Program(
            inputs=inputs,
            outputs=outputs,
            name="math_agent",
            description="A math agent",
        )

        input_query = Query(query="How much is 152648 + 485?")
        response = await agent(input_query)

        print(response.prettify_json())

    if __name__ == "__main__":
        asyncio.run(main())
    ```

    Result:

    ```json
    {
        "query": "How much is 152648 + 485?",
        "messages": [
            {
            "role": "assistant",
            "content": "Performing simple addition",
            "tool_calls": [
                {
                    "id": "92a3657c-1a45-46e6-8173-df4255b8423b",
                    "name": "calculate",
                    "arguments": {
                        "expression": "152648 + 485"
                        }
                    }
                ]
            },
            {
                "role": "tool",
                "content": {
                    "result": 153133.0,
                    "log": "Successfully executed"
                },
                "tool_call_id": "92a3657c-1a45-46e6-8173-df4255b8423b",
            },
            {
                "role": "assistant",
                "content": "The user has asked for a simple addition calculation. The assistant used the 'calculate' tool to perform this task, and the result was returned successfully.",
            }
        ],
        "final_answer": 153133.0
    }
    ```

    In *non-autonomous* mode (also called human in the loop or interactive mode), the
    user needs to validate/edit the tool arguments and send it back to the agent. In this
    mode, the agent requires an `ChatMessages` data model as input and output an
    `ChatMessage` (or `ChatMessages` if `return_inputs_with_trajectory` is true)
    back to the user. In that case, the agent ignore the `max_iterations` argument,
    as it will only perform one **step at a time**.

    Example:

    ```python
    import synalinks
    import asyncio

    MAX_ITERATIONS = 5

    async def calculate(expression: str):
        \"""Calculate the result of a mathematical expression.

        Args:
            expression (str): The mathematical expression to calculate, such as
                '2 + 2'. The expression can contain numbers, operators (+, -, *, /),
                parentheses, and spaces.
        \"""
        if not all(char in "0123456789+-*/(). " for char in expression):
            return {
                "result": None,
                "log": (
                        "Error: invalid characters in expression. "
                        "The expression can only contain numbers, operators (+, -, *, /),"
                        " parentheses, and spaces NOT letters."
                    ),
            }
        try:
            # Evaluate the mathematical expression safely
            result = round(float(eval(expression, {"__builtins__": None}, {})), 2)
            return {
                "result": result,
                "log": "Successfully executed",
            }
        except Exception as e:
            return {
                "result": None,
                "log": f"Error: {e}",
            }

    async def main():

        language_model = synalinks.LanguageModel(
            model="ollama/mistral",
        )

        tools = [
            synalinks.Tool(calculate),
        ]

        inputs = synalinks.Input(data_model=synalinks.ChatMessages)
        outputs = await synalinks.FunctionCallingAgent(
            tools=tools,
            language_model=language_model,
            return_inputs_with_trajectory=True,
            autonomous=False,
        )(inputs)
        agent = synalinks.Program(
            inputs=inputs,
            outputs=outputs,
            name="math_agent",
            description="A math agent",
        )

        input_messages = synalinks.ChatMessages(
            messages=[
                synalinks.ChatMessage(
                    role="user",
                    content="How much is 152648 + 485?",
                )
            ]
        )

        for i in range(MAX_ITERATIONS):

            response = await agent(input_messages)

            print("Assistant response (with trajectory):")
            print(response.prettify_json())

            assistant_message = response.get("messages")[-1]

            if not assistant_message.get("tool_calls"):
                break # We stop the loop if the agent didn't call any tool

            # Validate the tool calls arguments (with an UI or CLI)
            # Then re-inject the validated assistant response in the input_messages
            # The corresponding tools will be called by the agent
            # Here we assume everything is okay for the purpose of the demo.

            input_messages.messages.append(assistant_message)

    if __name__ == "__main__":
        asyncio.run(main())
    ```

    The FunctionCallingAgent is compatible with MCP tools,
    here is an example on how to use it:

    ```python
    import synalinks
    import asyncio
    import litellm

    class Query(synalinks.DataModel):
        \"""Input query data model\"""

        query: str = synalinks.Field(
            description="The user query",
        )

    class FinalAnswer(synalinks.DataModel):
        \"""Final answer data model\"""

        answer: str = synalinks.Field(
            description="The correct final answer",
        )


    async def main():

        language_model = synalinks.LanguageModel(
            model="ollama/mistral",
        )

        mcp_client = synalinks.MultiServerMCPClient(
            {
                "math": {
                    "url": "http://localhost:8183/mcp/",
                    "transport": "streamable_http",
                },
            }
        )

        tools = await mcp_client.get_tools()

        inputs = synalinks.Input(data_model=Query)
        outputs = await synalinks.FunctionCallingAgent(
            data_model=FinalAnswer,
            tools=tools,
            language_model=language_model,
            max_iterations=5,
            autonomous=True,
        )(inputs)

        agent = synalinks.Program(
            inputs=inputs,
            outputs=outputs,
            name="mcp_math_agent",
            description="A math agent that can use an external calculator",
        )

        input_query = Query(query="How much is 152648 + 485?")
        response = await agent(input_query)

        print(response.prettify_json())


    if __name__ == "__main__":
        asyncio.run(main())
    ```

    """

    def __init__(
        self,
        schema=None,
        data_model=None,
        language_model=None,
        prompt_template=None,
        examples=None,
        instructions=None,
        use_inputs_schema=False,
        use_outputs_schema=False,
        tools=None,
        autonomous=True,
        return_inputs_with_trajectory=True,
        max_iterations=5,
        name=None,
        description=None,
    ):
        super().__init__(
            name=name,
            description=description,
        )
        if not schema and data_model:
            schema = data_model.get_schema()
        self.schema = schema

        self.prompt_template = prompt_template

        if not instructions:
            instructions = get_default_instructions()
        self.instructions = instructions

        self.examples = examples
        self.use_inputs_schema = use_inputs_schema
        self.use_outputs_schema = use_outputs_schema
        self.language_model = language_model

        self.tools = {}
        if not tools:
            raise ValueError("You must set the `tools` argument")
        for tool in tools:
            self.tools[tool.name] = tool
        tool_calls_schema = dynamic_tool_calls(tools=tools)

        self.autonomous = autonomous
        self.return_inputs_with_trajectory = return_inputs_with_trajectory
        self.max_iterations = max_iterations

        self.tool_calls_generator = ChainOfThought(
            schema=tool_calls_schema,
            prompt_template=self.prompt_template,
            examples=self.examples,
            instructions=self.instructions,
            use_inputs_schema=self.use_inputs_schema,
            use_outputs_schema=self.use_outputs_schema,
            language_model=self.language_model,
            name=self.name + "_tool_calls_generator",
        )

        if self.schema and self.autonomous:
            self.final_generator = ChainOfThought(
                schema=self.schema,
                language_model=self.language_model,
                instructions=self.instructions,
                return_inputs=self.return_inputs_with_trajectory,
                name=self.name + "_final_generator",
            )

    async def call(self, inputs, training=False):
        if not inputs:
            return None
        if self.autonomous:
            if not is_chat_messages(inputs):
                trajectory = await ops.concat(
                    inputs,
                    ChatMessages(),
                    name=self.name + "_trajectory",
                )
            else:
                trajectory = inputs
        else:
            if not is_chat_messages(inputs):
                raise ValueError(
                    "In interactive mode, the FunctionCallingAgent "
                    "needs an ChatMessages-like data model as inputs"
                )
            trajectory = inputs

        agent_messages = trajectory.get("messages")

        if self.autonomous:
            for i in range(self.max_iterations):
                tool_calls = await self.tool_calls_generator(trajectory)

                if not tool_calls:
                    assistant_message = ChatMessage(
                        role=ChatRole.ASSISTANT,
                        content="Something went wrong while trying to decide the next action.",
                    )
                    agent_messages.append(assistant_message.get_json())
                    break

                assistant_message = ChatMessage(
                    role=ChatRole.ASSISTANT,
                    content=tool_calls.get("thinking", ""),
                )

                if not tool_calls.get("tool_calls"):
                    agent_messages.append(assistant_message.get_json())
                    break

                tasks = []
                tool_calls_ids = []

                for tool_call in tool_calls.get("tool_calls"):
                    tool_name = tool_call.get("tool_name")
                    tools_arguments = out_mask_json(tool_call, mask=["tool_name"])
                    tool_call_id = str(uuid.uuid4())
                    tool_calls_ids.append(tool_call_id)
                    assistant_message.tool_calls.append(
                        ToolCall(
                            id=tool_call_id,
                            name=tool_name,
                            arguments=tools_arguments,
                        )
                    )
                    tasks.append(self.tools[tool_name](**tools_arguments))

                agent_messages.append(assistant_message.get_json())

                tool_results = await asyncio.gather(*tasks, return_exceptions=True)
                for j, tool_result in enumerate(tool_results):
                    tool_call_id = tool_calls_ids[j]
                    if isinstance(tool_result, Exception):
                        agent_messages.append(
                            ChatMessage(
                                role=ChatRole.TOOL,
                                tool_call_id=tool_call_id,
                                content="error: %s" % str(tool_result),
                            ).get_json()
                        )
                    else:
                        agent_messages.append(
                            ChatMessage(
                                role=ChatRole.TOOL,
                                tool_call_id=tool_call_id,
                                content=tool_result,
                            ).get_json()
                        )

                trajectory.update({"messages": agent_messages})
            if self.schema:
                return await self.final_generator(trajectory)
            else:
                return trajectory
        else:
            if len(agent_messages) > 0:
                if agent_messages[-1].get("role") == ChatRole.ASSISTANT:
                    tasks = []
                    tool_calls_ids = []

                    tool_calls = agent_messages[-1].get("tool_calls")
                    for tool_call in tool_calls:
                        tool_name = tool_call.get("name")
                        tools_arguments = tool_call.get("arguments")
                        tool_call_id = tool_call.get("id")
                        tool_calls_ids.append(tool_call_id)
                        tasks.append(self.tools[tool_name](**tools_arguments))

                    tool_results = await asyncio.gather(*tasks, return_exceptions=True)
                    for j, tool_result in enumerate(tool_results):
                        tool_call_id = tool_calls_ids[j]
                        if isinstance(tool_result, Exception):
                            agent_messages.append(
                                ChatMessage(
                                    role=ChatRole.TOOL,
                                    tool_call_id=tool_call_id,
                                    content="error: %s" % str(tool_result),
                                ).get_json()
                            )
                        else:
                            agent_messages.append(
                                ChatMessage(
                                    role=ChatRole.TOOL,
                                    tool_call_id=tool_call_id,
                                    content=tool_result,
                                ).get_json()
                            )

            trajectory.update({"messages": agent_messages})

            tool_calls = await self.tool_calls_generator(trajectory)

            assistant_message = ChatMessage(
                role=ChatRole.ASSISTANT,
                content=tool_calls.get("thinking", ""),
                tool_calls=[],
            )

            for tool_call in tool_calls.get("tool_calls", []):
                tool_name = tool_call.get("tool_name")
                tools_arguments = out_mask_json(tool_call, mask=["tool_name"])
                tool_call_id = str(uuid.uuid4())

                assistant_message.tool_calls.append(
                    ToolCall(
                        id=tool_call_id,
                        name=tool_name,
                        arguments=tools_arguments,
                    )
                )

            agent_messages.append(assistant_message.get_json())
            trajectory.update({"messages": agent_messages})

            if self.return_inputs_with_trajectory:
                return JsonDataModel(
                    json=trajectory.get_json(),
                    schema=ChatMessages.get_schema(),
                    name=self.name,
                )
            else:
                return JsonDataModel(
                    json=assistant_message.get_json(),
                    schema=ChatMessage.get_schema(),
                    name=self.name,
                )

    async def compute_output_spec(self, inputs, training=False):
        if self.autonomous:
            _ = await self.tool_calls_generator(inputs)
            if self.schema:
                return await self.final_generator(inputs)
            else:
                return SymbolicDataModel(
                    schema=ChatMessages.get_schema(),
                    name=self.name,
                )
        else:
            if not is_chat_messages(inputs):
                raise ValueError(
                    "In interactive mode, the FunctionCallingAgent "
                    "needs an ChatMessages-like data model as inputs"
                )

            _ = await self.tool_calls_generator(inputs)

            if self.return_inputs_with_trajectory:
                return SymbolicDataModel(
                    schema=ChatMessages.get_schema(),
                    name=self.name,
                )
            else:
                return SymbolicDataModel(
                    schema=ChatMessage.get_schema(),
                    name=self.name,
                )

    def get_config(self):
        config = {
            "schema": self.schema,
            "prompt_template": self.prompt_template,
            "examples": self.examples,
            "instructions": self.instructions,
            "use_inputs_schema": self.use_inputs_schema,
            "use_outputs_schema": self.use_outputs_schema,
            "autonomous": self.autonomous,
            "max_iterations": self.max_iterations,
            "return_inputs_with_trajectory": self.return_inputs_with_trajectory,
            "name": self.name,
            "description": self.description,
        }
        language_model_config = {
            "language_model": serialization_lib.serialize_synalinks_object(
                self.language_model,
            )
        }
        tools_config = {
            "tools": [
                serialization_lib.serialize_synalinks_object(tool)
                for tool in self.tools.values()
            ]
        }
        return {**config, **language_model_config, **tools_config}

    @classmethod
    def from_config(cls, config):
        tools = [
            serialization_lib.deserialize_synalinks_object(tool)
            for tool in config.pop("tools")
        ]
        language_model = serialization_lib.deserialize_synalinks_object(
            config.pop("language_model")
        )
        return cls(
            language_model=language_model,
            tools=tools,
            **config,
        )



# Modified from: keras/src/ops/node.py
# Original authors: François Chollet et al. (Keras Team)
# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import collections

from synalinks.src import tree
from synalinks.src.backend import SymbolicDataModel
from synalinks.src.ops.symbolic_arguments import SymbolicArguments


class Node:
    """A `Node` describes an operation `__call__()` event.

    A Synalinks Function is a DAG with `Node` instances as nodes, and
    `SymbolicDataModel` instances as edges. Nodes aren't `Operation` instances,
    because a single operation could be called multiple times, which would
    result in graph cycles.

    A `__call__()` event involves input data_models (and other input arguments),
    the operation that was called, and the resulting output data_models.
    A `Node` will include all this information.

    Since a single `Operation` could be called multiple times,
    the `Node` instances are stored on operations as a list.
    Each time an operation is called, a node is added to `op._inbound_nodes`.
    Each time the output of an operation is used by another operation,
    a node is added to `op._outbound_nodes`.

    Every `SymbolicDataModel` instance has a `synalinksHistory` object attached,
    which tracks the `Node` that records the `__call__()` event that created
    the data_model. By recursively walking through `Node` instances
    via the `KerasHistory` metadata of `SymbolicDataModel` instances, once can
    retrieve the entire DAG of a synalinks Function.

    Args:
        operation (Operation): The Operation that was called in the `op.__call__()`
            event that this node represents.
        call_arg (positional arguments): The positional arguments the operation was
            called with.
        call_kwargs (keyword arguments): The keyword arguments the operation was
            called with.
        outputs (list): The output data models of the `op.__call__()` call.
    """

    def __init__(self, operation, call_args=None, call_kwargs=None, outputs=None):
        self.operation = operation
        self.arguments = SymbolicArguments(*call_args, **call_kwargs)
        self.outputs = [] if outputs is None else tree.flatten(outputs)
        for x in self.outputs:
            if not isinstance(x, SymbolicDataModel):
                raise ValueError(
                    "All operation outputs must be data_models. "
                    f"Operation {operation} returned a non-data_model. "
                    f"Non-data_model received: {x}"
                )

        zero_history = any(
            not x.record_history for x in self.arguments.symbolic_data_models
        )

        # If inputs don't have metadata yet, add it.
        if not zero_history:
            for data_model in self.arguments.symbolic_data_models:
                if not hasattr(data_model, "_synalinks_history"):
                    data_model._synalinks_history = SynalinksHistory(
                        operation=None, node_index=0, data_model_index=0
                    )

        # Wire up Node to Operations.
        self.operation._inbound_nodes.append(self)
        for dt in self.arguments.symbolic_data_models:
            inbound_op = dt._synalinks_history.operation
            if inbound_op is not None:  # It's a graph entry point.
                inbound_op._outbound_nodes.append(self)

        # Set metadata on outputs.
        if not zero_history:
            node_index = len(self.operation._inbound_nodes) - 1
            for i, data_model in enumerate(self.outputs):
                data_model._synalinks_history = SynalinksHistory(
                    operation=operation,
                    node_index=node_index,
                    data_model_index=i,
                )

        # Whether this is a root node.
        self.is_input = not self.arguments.symbolic_data_models

    def __repr__(self):
        if self.operation.description:
            operation_str = f"({self.operation.name}, '{self.operation.description}')"
        else:
            operation_str = f"{self.operation.name}"
        return f"<Node operation={operation_str}, id={id(self)}>"

    @property
    def input_data_models(self):
        return self.arguments.symbolic_data_models

    @property
    def output_data_models(self):
        return self.outputs

    @property
    def parent_nodes(self):
        """The parent `Node`s.

        Returns:
            all the `Node`s whose output this node immediately depends on.
        """
        node_deps = []
        for dt in self.arguments.symbolic_data_models:
            op = dt._synalinks_history.operation
            node_index = dt._synalinks_history.node_index
            if op is not None:  # `None` for `Input` data_models.
                node_deps.append(op._inbound_nodes[node_index])
        return node_deps


class SynalinksHistory(
    collections.namedtuple(
        "SynalinksHistory", ["operation", "node_index", "data_model_index"]
    )
):
    """Tracks the Operation call that created a DataModel.

    During construction of synalinks Functions, this metadata is added to
    each DataModel produced as the output of an Operation.
    This allows synalinks to track how each DataModel was produced, and
    this information is later retraced by the `Function` class to
    reconstruct the Operations graph.

    Attributes:
      operation: The Operation instance that produced the DataModel.
      node_index: The specific call to the Operation that produced this DataModel.
        Operations can be called multiple times in order to share weights. A new
        node is created every time an Operation is called. The corresponding
        node that represents the call event that produced the DataModel can be
        found at `op._inbound_nodes[node_index]`.
      data_model_index: The output index for this DataModel.
        Always zero if the Operation that produced this DataModel
        only has one output.
    """

    # Added to maintain memory and performance characteristics of `namedtuple`
    # while subclassing.
    __slots__ = ()


def is_symbolic_data_model(obj):
    return hasattr(obj, "_synalinks_history")



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)


import numpy as np
from datasets import load_dataset

from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import DataModel
from synalinks.src.backend import Entity


class Document(Entity):
    title: str
    text: str


class Question(DataModel):
    question: str


class Answer(DataModel):
    answer: str


@synalinks_export("synalinks.datasets.hotpotqa.get_knowledge_data_model")
def get_knowledge_data_model():
    """
    Returns HotpotQA knowledge data model for pipeline configurations.

    Returns:
        (DataModel): The HotpotQA knowledge data model
    """
    return Document


@synalinks_export("synalinks.datasets.hotpotqa.get_input_data_model")
def get_input_data_model():
    """
    Returns HotpotQA input data model for pipeline configurations.

    Returns:
        (DataModel): The HotpotQA input data model
    """
    return Question


@synalinks_export("synalinks.datasets.hotpotqa.get_output_data_model")
def get_output_data_model():
    """
    Returns HotpotQA output data model for pipeline configurations.

    Returns:
        (DataModel): The HotpotQA output data model
    """
    return Answer


@synalinks_export("synalinks.datasets.hotpotqa.load_knowledge")
def load_knowledge():
    """
    Load and format data from HuggingFace

    Example:

    ```python
    knowledge = synalinks.datasets.hotpotqa.load_knowledge()
    ```

    Returns:
        (list): The  data ready for knowledge injestion
    """
    documents = []
    train_examples = load_dataset(
        "hotpot_qa", "fullwiki", split="train", trust_remote_code=True
    )
    for raw_example in train_examples:
        context = raw_example.get("context", None)
        if context:
            for i in range(len(context["title"])):
                documents.append(
                    Document(
                        title=context["title"][i],
                        text="\n".join(context["sentences"][i]),
                    )
                )
    documents = np.array(documents, dtype="object")
    return documents


@synalinks_export("synalinks.datasets.hotpotqa.load_data")
def load_data():
    """
    Load and format data from HuggingFace

    Example:

    ```python
    (x_train, y_train), (x_test, y_test) = synalinks.datasets.hotpotqa.load_data()
    ```

    Returns:
        (tuple): The train and test data ready for training
    """
    x_train = []
    y_train = []
    x_test = []
    y_test = []

    train_examples = load_dataset(
        "hotpot_qa", "fullwiki", split="train", trust_remote_code=True
    )
    eval_examples = load_dataset(
        "hotpot_qa", "fullwiki", split="validation", trust_remote_code=True
    )

    for raw_example in train_examples:
        x_train.append(Question(question=raw_example["question"]))
        y_train.append(Answer(answer=raw_example["answer"]))

    for raw_example in eval_examples:
        if raw_example["level"] == "hard":
            x_test.append(Question(question=raw_example["question"]))
            y_test.append(Answer(answer=raw_example["answer"]))

    x_train = np.array(x_train, dtype="object")
    y_train = np.array(y_train, dtype="object")

    x_test = np.array(x_test, dtype="object")
    y_test = np.array(y_test, dtype="object")

    return (x_train, y_train), (x_test, y_test)
